# coding=utf-8
"""
Abstract class allowing for torch.nn.Modules to also host a cache of torch objects
which can also be ported like torch.nn.Parameters are, on GPU for example.
"""
from __future__ import annotations

import torch
from typing import Union, List, Iterable, Any, Type
from abc import ABCMeta

from .module import Module
from ..utils import reverse_dict, extend_docstring, DEFAULT_CACHE_LEVEL


@extend_docstring(Module)
class Cache(Module,
            metaclass=ABCMeta):
    r"""
    :param cache_level: Cache level for saving temporary execution results during the execution. The higher the cache,
        the more is saved. Defaults to ``'normal'``. We refer to the :doc:`/features/cache` documentation for further
        information.
    :type cache_level: str, optional
    """

    _cache_elements = []

    _cache_level_switcher = {"none": 0,
                             "light": 1,
                             "normal": 2,
                             "heavy": 3,
                             "total": 4}

    def __init__(self, *args, **kwargs):
        super(Cache, self).__init__(*args, **kwargs)

        # we initiate the cache
        self._cache = {}
        self.cache_level = kwargs.pop('cache_level', 'normal')

    @property
    def cache_level(self) -> str:
        r"""
        Cache level for saving temporary execution results during the execution. The higher the cache,
        the more is saved. Defaults to ``'normal'`` unless set otherwise during instantiation. The different possible
        values are:

        * ``"none"``: the cache is non-existent and everything is computed on the go.
        * ``"light"``: the cache is very light. For example, only the kernel matrix and statistics of the sample points are saved.
        * ``"normal"``: same as light, but the statistics of the out-of-sample points are also saved.
        * ``"heavy"``: in addition to the statistics, the final kernel matrices of the out-of-sample points are saved.
        * ``"total"``: every step of any computation is saved.

        We refer to the :doc:`/features/cache` documentation for further information.
        """
        switcher = Cache._cache_level_switcher
        inv_switcher = {switcher[k]: k for k in switcher}
        return inv_switcher[self._cache_level]

    @cache_level.setter
    def cache_level(self, val: Union[str, int]):
        val = self._get_level(val)
        self._cache_level = val
        for value in self._cache.values():
            if isinstance(value, Cache):
                value._cache_level = val

    def _get_level(self, level: Union[int, str, None]) -> int:
        r"""
        Transforms the cache level to an int if not already.
        """
        if level is None:
            return self._cache_level
        if isinstance(level, str):
            level = Cache._cache_level_switcher.get(level, "Unrecognized cache level.")
        return level

    def _apply(self, fn, recurse=True):
        r"""
        This if the native function by :external+torch:py:class:`torch.nn.modules.module.Module`, used when porting the
        module. This ensures that the cache is also ported. This is used for example to port the data to the GPU or CPU.

        .. note::
            This method is documented for completeness, but it should never be required to call it directly.

        """
        with torch.no_grad():
            for value in self._cache.values():
                cache_entry = value[2]
                if isinstance(cache_entry, torch.Tensor):
                    cache_entry.data = fn(cache_entry)
                elif isinstance(cache_entry, Cache):
                    cache_entry._apply(fn)
        if recurse:
            for child in self.children():
                child._apply(fn, recurse=recurse)
        return super(Cache, self)._apply(fn)

    def _save(self, key, fun, level_key=None, default_level: str = 'total', force: bool = False,
              persisting=False) -> Any:
        r"""
        Saves an element in the cache.

        :param key: key of the cache element.
        :param fun: function to compute the element if not in the cache already.
        :param level_key: key referencing the default level to use in ``kerch.DEFAULT_CACHE_LEVELS``. If not specified, the ``default_level`` argument is used.
        :param default_level: level where to save the cache element. Defaults to 'total'.
        :param force: if the value is ``True``, the element will nevertheless be saved whatever level is specified. Defaults to ``False``.
        :param persisting: These values are meant to persist after a cache reset when calling
            :py:meth:`~kerch.feature.Cache.reset` with ``reset_persisting=False``. Defaults to ``False``.

        :type key: str
        :type fun: function handle
        :type level_key: str, optional
        :type default_level: str, optional
        :type force: bool, optional
        :type persisting: bool, optional
        :return: The result of ``fun()``
        """

        # determine which level to store in
        try:
            level = DEFAULT_CACHE_LEVEL[level_key]
        except KeyError:
            level = default_level
        level = self._get_level(level)

        # we now compute, store if the level corresponds and return the value
        assert callable(fun) is not None, \
            f"Cannot store {key} in the cache as no callable argument fun has been provided"
        val = fun()
        if level <= self._cache_level or force:
            self._cache[key] = (level, persisting, val)
        return val

    def _get(self, key, fun=None, level_key=None, default_level: str = 'normal', force: bool = False,
             overwrite: bool = False, persisting=False, destroy=False) -> Any:
        r"""
        Retrieves an element from the cache. If the element is not present, it saved to the cache provided its level
        is lower or equal to the default level. This can be overwritten by the overwrite argument.

        :param key: key of the cache element.
        :param fun: function to compute the element if not in the cache already.
        :param level_key: key referencing the default level to use in ``kerch.DEFAULT_CACHE_LEVELS``. If not specified, the ``default_level`` argument is used.
        :param default_level: level where to save the cache element. Defaults to 'normal'.
        :param force: if the value is ``True``, the element will nevertheless be saved whatever level is specified. Defaults to ``False``.
        :param persisting: These values are meant to persist after a cache reset when calling
            :py:meth:`~kerch.feature.Cache.reset` with ``reset_persisting=False``. Defaults to ``False``.
        :param destroy: This destroys the value from the cache after being read/computed. This is meant for short-term
            memory. Defaults to ``False``.

        :type key: str
        :type level_key: str, optional
        :type default_level: str, optional
        :type fun: function handle
        :type force: bool, optional
        :type overwrite: bool, optional
        :type persisting: bool, optional
        :type destroy: bool, optional

        :return: The result of ``fun()``
        """

        # a cache element is represented by the tuple(level, persisting, value)

        # we first check if the value is already in the cache
        if not overwrite:
            if key in self._cache:
                return self._cache[key][2]
            # elif type(key) is tuple:
            #     reverted_key = (key[1], key[0])
            #     if reverted_key in self._cache:
            #         return self._cache[reverted_key][1].T

        # if it is not, we save it
        val = self._save(key=key, fun=fun, level_key=level_key, default_level=default_level,
                         force=force, persisting=persisting)
        if destroy:
            self._remove_from_cache(key)
        return val

    def _reset_cache(self, reset_persisting: bool = True, avoid_classes: list | None = None) -> None:
        r"""
        This just resets the cache and makes it empty.

        :param reset_persisting: Persisting elements are meant to resist to a cache reset
            (see :py:meth:`~kerch.feature.Cache._save`). The option allows to also reset them if ``True``. Defaults to ``True``.
        :type reset_persisting: bool, optional
        :param avoid_classes: Class of which the elements must be avoided to be resetted. Default to ``[]``.
        :type avoid_classes: list(type(:class:`~kerch.feature.Cache`)), optional

        .. note::
            This method is documented for completeness, but it should never be required to call it directly.

        """
        elements_avoided = list()
        if avoid_classes is not None:
            for cl in avoid_classes:
                elements_avoided.extend(cl._cache_elements)

        if reset_persisting and len(elements_avoided) == 0:
            self._logger.debug("The cache is fully resetted.")
            self._cache = {}
        else:
            self._logger.debug("The cache is resetted at the exception of the persisting elements and avoided classes.")
            for key in list(self._cache.keys()):
                if key in self._cache and key not in elements_avoided:
                    val = self._cache[key]
                    if not val[1] or reset_persisting:
                        self._cache.pop(key)

    def _clean_cache(self, max_level: Union[str, int, None] = None):
        r"""
        Cleans all cache elements above a certain level. This is relevant for cleaning the cache elements that have
        been forced (see :py:meth:`~kerch.feature.Cache._save`).

        :param max_level: all levels above this level will be cleaned, ``max_level`` excluded. Defaults to the default
            cache level.
        :type max_level: str | int, optional

        .. note::
            This method is documented for completeness, but it should never be required to call it directly.

        """
        max_level = self._get_level(max_level)
        self._logger.debug(f"The cache is cleaned for levels {max_level} and above).")
        # for val in self._cache.values(): del val
        if max_level == 0:
            self._reset_cache()
        else:
            for key in list(self._cache.keys()):
                if key in self._cache.keys():
                    value = self._cache[key]
                    if value[0] > max_level:
                        self._cache.pop(key)

    def _remove_from_cache(self, key: Union[str, List[str]]) -> None:
        r"""
        Removes one or more specific element(s) from the cache.

        :param key: Key(s) of the cache element to be removed.
        :type key: str | list[str]

        .. note::
            This method is documented for completeness, but it should never be required to call it directly.

        """

        def _remove_entry(key_item) -> None:
            if key_item in self._cache:
                self._cache.pop(key_item)

        if not hasattr(key, '__iter__'):
            key = [key]

        for key_item in key:
            _remove_entry(key_item)

    def reset(self, recurse=False, reset_persisting=True) -> None:
        r"""
        Resets the cache to be empty. We refer to the :doc:`/features/cache` documentation for more information.

        :param recurse: If ``True``, resets the cache of this module and also of its potential children. otherwise,
            it only resets the cache for this module. Defaults to ``True``.
        :type recurse: bool, optional
        :param reset_persisting: Persisting elements are meant to resist to a cache reset (see
            :py:meth:`~kerch.feature.Cache._save`). The option allows to also reset them if ``True``. Defaults to
            ``True``.
        :type reset_persisting: bool, optional
        """
        self._reset_cache(reset_persisting=reset_persisting)
        if recurse:
            for cache in self.children():
                if isinstance(cache, Cache):
                    cache.reset(recurse=recurse, reset_persisting=reset_persisting)

    def cache_keys(self, private: bool = False) -> Iterable[str]:
        r"""
        Returns an iterable containing the different cache keys.
        We refer to the :doc:`/features/cache` documentation for more information.

        :param private: Some cache elements are private and are not returned unless set to ``True``. Defaults to ``False``.
        :type private: bool, optional
        """
        for key in self._cache.keys():
            if key[0] != "_" or private:
                yield key

    def print_cache(self, private: bool = False) -> None:
        r"""
        Prints the cache content. We refer to the :doc:`/features/cache` documentation for further information.

        :param private: Some cache elements are private and are not returned unless set to ``True``. Defaults to ``False``.
        :type private: bool, optional
        """
        from ..transform import TransformTree
        transforms = list()

        for key, val in self._cache.items():
            if key[0] != "_" or private:
                if isinstance(val[2], TransformTree):
                    transforms.append((key, val[2]))
                print(key, end=' [')
                level = reverse_dict(Cache._cache_level_switcher).get(val[0])
                print(level, end='] ')
                if val[1]:
                    print('persisting', end='')
                print(end=': ')
                print(val[2])

        for key, transform in transforms:
            print('\n' + key.upper() + ":")
            transform.print_cache(private)



# coding=utf-8
"""
File containing the linear kernel class.

@author: HENRI DE PLAEN
@copyright: KU LEUVEN
@license: MIT
@date: March 2021
"""

import torch
from .. import utils
from .kernel import Kernel
from abc import ABCMeta, abstractmethod


@utils.extend_docstring(Kernel)
class Explicit(Kernel, metaclass=ABCMeta):

    @utils.kwargs_decorator({})
    def __init__(self, *args, **kwargs):
        super(Explicit, self).__init__(*args, **kwargs)
        self._dim_feature = None

    def __str__(self):
        return f"Explicit kernel."

    @property
    def explicit(self) -> bool:
        return True

    @property
    def dim_feature(self) -> int:
        if self._dim_feature is None:
            # if it has not been set before, we can compute it with a minimal example
            self._dim_feature = self._explicit_with_none(x=self.current_sample_projected[0:1, :]).shape[1]
        return self._dim_feature

    def _implicit(self, x, y):
        phi_x = self._explicit(x)
        phi_y = self._explicit(y)
        return phi_x @ phi_y.T

    @abstractmethod
    def _explicit(self, x):
        return x

    def _explicit_preimage(self, phi):
        raise NotImplementedError





# coding=utf-8
from __future__ import annotations
from math import sqrt
import torch
from torch import Tensor as T
from abc import ABCMeta
from typing import Union, Iterator

from ._Level import _Level
from .. import utils


class _KPCA(_Level, metaclass=ABCMeta):
    r"""
    Kernel Principal Component Analysis.

    :param prune_small_vals: Indicates whether the eigenvalues smaller than the machine precision should be pruned.
    """

    @utils.extend_docstring(_Level)
    def __init__(self, *args, **kwargs):
        super(_KPCA, self).__init__(*args, **kwargs)
        self._vals = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE),
                                        requires_grad=False)
        self._subloss_projected = None
        self._subloss_original = None

        self._prune_small_vals = kwargs.pop('prune_small_vals', False)

    @property
    def vals(self) -> T:
        r"""
        Eigenvalues of the model. The model has to be fitted for these values to exist.
        """
        return self._vals

    @property
    def sqrt_vals(self) -> T:
        return self._get(key="sqrt_vals", fun=lambda: torch.sqrt(self.vals))

    @vals.setter
    def vals(self, val):
        val = utils.castf(val, tensor=False, dev=self._vals.device)
        self._vals.data = val

    def total_variance(self, as_tensor=False, normalize=True, representation=None) -> Union[float, T]:
        r"""
        Total variance contained in the feature map. In primal formulation,
        this is given by :math:`\DeclareMathOperator{\tr}{tr}\tr(C)`, where :math:`C = \sum\phi(x)\phi(x)^\top` is
        the covariance matrix on the sample. In dual, this is given by :math:`\DeclareMathOperator{\tr}{tr}\tr(K)`,
        where :math:`K_{ij} = k(x_i,x_j)` is the kernel matrix on the sample.

        :param as_tensor: Indicated whether the variance has to be returned as a float or a torch.Tensor., defaults
            to ``False``
        :type as_tensor: bool, optional

        .. warning::
            For this value to strictly be interpreted as a variance, the corresponding kernel (or feature map)
            has to be normalized. In that case however, the total variance will amount to the dimension of the feature
            map in primal and the number of datapoints in dual.
        """
        representation = utils.check_representation(representation=representation, default=self._representation)
        level_key = "KPCA_total_variance_default_representation" if representation == self._representation \
            else "KPCA_total_variance_other_representation"
        if representation == 'primal':
            var = self._get("total_variance_primal", level_key=level_key, fun=lambda: torch.trace(self.C))
        else:
            var = self._get("total_variance_dual", level_key=level_key, fun=lambda: torch.trace(self.K))
        if normalize:
            var /= self.num_idx
        if as_tensor:
            return var
        return var.detach().cpu().numpy()

    def model_variance(self, as_tensor=False, normalize=True) -> Union[float, T]:
        r"""
        Total variance learnt by the model given by the sum of the eigenvalues.

        :param as_tensor: Indicated whether the variance has to be returned as a float or a torch.Tensor., defaults
            to ``False``
        :type as_tensor: bool, optional

        .. warning::
            For this value to strictly be interpreted as a variance, the corresponding kernel (or feature map)
            has to be normalized. We refer to the remark of ``total_variance``.
        """
        var = torch.sum(self.vals)
        if normalize:
            var /= self.num_idx
        if as_tensor:
            return var
        return var.detach().cpu().numpy()

    def _reset_dual(self) -> None:
        super(_KPCA, self)._reset_dual()
        self._remove_from_cache(["total_variance_primal", "total_variance_dual"])

    def relative_variance(self, as_tensor=False) -> Union[float, T]:
        r"""
        Relative variance learnt by the model given by ```model_variance``/``total_variance``.
        This number is always comprised between 0 and 1 and avoids any considerations on normalization.

        :param as_tensor: Indicated whether the variance has to be returned as a float or a torch.Tensor., defaults
            to ``False``
        :type as_tensor: bool, optional
        """
        var = self.model_variance(as_tensor=as_tensor, normalize=False) / \
              self.total_variance(as_tensor=as_tensor, normalize=False)
        return var

    def _solve_primal(self) -> None:
        C = self.C

        if self._dim_output is None:
            self._dim_output = self.dim_feature
            self._logger.warning(f"The output dimension has not been set and is now set to its maximum possible "
                              f"value. In primal formulation, this corresponds to the feature dimension "
                              f"(dim_output=dim_feature={self.dim_output}).")
        elif self._dim_output > self.dim_feature:
            self._logger.warning(f"In primal, the output dimension {self.dim_output} (the number of "
                              f"eigenvectors) must not exceed the feature dimension {self.dim_feature} (the dimension "
                              f"of the correlation matrix to be diagonalized). As this is the case here, the output "
                              f"dimension is reduced to {self.dim_feature}.")
            self.dim_output = self.dim_feature

        v, e = utils.eigs(C, k=self.dim_output, psd=True)

        # prune very small eigenvalues if they exist to avoid unstability due to the later inversion
        idx_small = v < utils.EPS
        sum_small = torch.sum(idx_small)
        if sum_small > 0:
            self._logger.warning(
                f"{sum_small} very small or negative eigenvalues are detected on {self._dim_output}. "
                f"To avoid numerical instability, these values are pruned. "
                f"The new explicit dimension is now {self._dim_output - sum_small}.")
            keep_idx = torch.logical_not(idx_small)
            v = v[keep_idx]
            e = e[:, keep_idx]
            self._dim_output = self._dim_output - sum_small

        self.primal_param = e
        self.vals = v

    def _solve_dual(self) -> None:
        K = self.K

        if self._dim_output is None:
            self._dim_output = self.num_idx
            self._logger.warning(f"The output dimension has not been set and is now set to its maximum possible "
                              f"value. In dual formulation, this corresponds to the number of datapoints in the "
                              f"current sample (dim_output=num_idx={self.dim_output}).")
        elif self._dim_output > self.num_idx:
            self._logger.warning(f"In dual, the output dimension {self.dim_output} (the number of "
                              f"eigenvectors) must not exceed the number of samples {self.num_idx} (the dimension "
                              f"of the kernel matrix to be diagonalized). As this is the case here, the output "
                              f"dimension is reduced to {self.num_idx}.")
            self.dim_output = self.num_idx

        v, e = utils.eigs(K, k=self.dim_output, psd=True)
        fact = 1 / self.num_idx

        # prune very small eigenvalues if they exist to avoid unstability due to the later inversion
        idx_small = v < utils.EPS
        sum_small = torch.sum(idx_small)
        if sum_small > 0:
            self._logger.warning(
                f"{sum_small} very small or negative eigenvalues are detected on {self._dim_output}. "
                f"To avoid numerical instability, these values are pruned. "
                f"The new explicit dimension is now {self._dim_output - sum_small}.")
            keep_idx = torch.logical_not(idx_small)
            v = v[keep_idx]
            e = e[:, keep_idx]
            self._dim_output = self._dim_output - sum_small

        self.update_dual(e)
        self.vals = fact * v

    @utils.extend_docstring(_Level.solve)
    @torch.no_grad()
    def solve(self, sample=None, target=None, representation=None, **kwargs) -> None:
        r"""
        Solves the model by decomposing the kernel matrix or the covariance matrix in principal components
        (eigendecomposition).
        """
        # KPCA models don't require the target to be defined. This is verified.
        if target is not None:
            self._logger.warning("The target value is not used when fitting a KPCA model.")
        return _Level.solve(self,
                            sample=sample,
                            target=None,
                            representation=representation)

    def _stiefel_parameters(self, recurse=True) -> Iterator[torch.nn.Parameter]:
        # the stiefel optimizer requires the first dimension to be the number of eigenvectors
        yield from super(_KPCA, self)._stiefel_parameters(recurse)
        if self._representation == 'primal':
            if self._primal_param_exists:
                yield self._primal_param
        else:
            if self._dual_param_exists:
                yield self._dual_param

    @property
    def H(self) -> T:
        return self.dual_param

    @property
    def W(self) -> T:
        return self.primal_param @ torch.diag(self.sqrt_vals)

    def loss(self, representation=None) -> T:
        r"""
        Reconstruction error on the sample.
        """
        return self._loss_original(representation=representation) \
            - self._loss_projected(representation=representation)

    def _loss_original(self, representation=None) -> T:
        representation = utils.check_representation(representation, self._representation, self)
        level_key = "Level_subloss_default_representation" if self._representation == representation \
            else "Level_subloss_representation"

        def fun():
            if representation == 'primal':
                M = self.C
            else:
                M = self.K
            return torch.trace(M)

        return self._get(key='subloss_original_' + representation,
                         level_key=level_key, fun=fun)

    def _loss_projected(self, representation=None) -> T:
        representation = utils.check_representation(representation, self._representation, self)
        level_key = "Level_subloss_default_representation" if self._representation == representation \
            else "Level_subloss_representation"

        def fun():
            if representation == 'primal':
                U = self._primal_param  # transposed compared to primal_param
                M = self.C
            else:
                U = self._dual_param  # transposed compared to dual_param
                M = self.K
            return torch.trace(U.T @ U @ M)

        return self._get(key='subloss_projected_' + representation,
                         level_key=level_key, fun=fun)

    def losses(self, representation=None) -> dict:
        return {'Original': self._loss_original().data.detach().cpu().item(),
                'Projected': self._loss_projected().data.detach().cpu().item(),
                **super(_KPCA, self).losses()}

    @torch.no_grad()
    def h(self, phi: T | None = None, k: T | None = None) -> T:
        r"""
        Draws a `h` given the maximum a posteriori of the distribution. By choosing the input, you either
        choose a primal or dual representation.

        :param phi: Primal representation.
        :param k: Dual representation.
        :type phi: Tensor[N, dim_input], optional
        :type k: Tensor[N, num_idx], optional
        :return: MAP of h given phi or k.
        :rtype: Tensor[N, dim_output]
        """
        if phi is not None and k is None:
            return phi @ self.W @ torch.diag(1 / self.vals)
        if phi is None and k is not None:
            return k @ self.H @ torch.diag(1 / self.vals)
        else:
            raise AttributeError("One and only one attribute phi or k has to be specified.")

    @torch.no_grad()
    def phi_map(self, h: T) -> T:
        r"""
        Feature representation :math:`\phi(x^\star)` given a latent representation :math:`h^\star`.

        .. math::
            \phi(x^\star) = = W h^\star.

        :param h: Latent representation :math:`h^\star`.
        :type h: Tensor[N, dim_output]
        :return: Feature representation :math:`\phi(x^\star)`.
        :rtype: Tensor[N, dim_feature]
        """
        return h @ self.W.T

    @torch.no_grad()
    def k_map(self, h: T) -> T:
        r"""
        RKHS representation :math:`k(x^\star,\mathtt{sample})` given a latent representation :math:`h^\star`.

        .. math::
            k(x^\star, x_j) = KH^\toph^\star,

        with :math:`K` the kernel matrix on the sample :py:attr:`self.K` and :math:`H` the hidden vectors :py:attr:`self.hidden`.

        :param h: Latent representation :math:`h^\star`.
        :type h: Tensor[N, dim_output]
        :return: RKHS representation :math:`k(x^\star,\mathtt{sample})`.
        :rtype: Tensor[N, num_idx]
        """
        return h @ self.H.T @ self.K

    @torch.no_grad()
    def draw_h(self, num: int = 1) -> T:
        r"""
        Draws a :math:`h^\star` normally.

        :param num: Number of :math:`h^\star` to be sampled, defaults to 1.
        :type num: int, optional
        :return: Latent representation.
        :rtype: torch.Tensor [num, dim_output]
        """
        return torch.randn((num, self.dim_output), device=self._dual_param.device, dtype=utils.FTYPE)

    @torch.no_grad()
    def draw_phi(self, num: int = 1, posterior: bool = True) -> T:
        r"""
        Draws a primal representation phi given its posterior distribution.

        :param posterior: Indicates whether phi has to be drawn from its posterior distribution or its conditional
            given the prior of h. Defaults to True.
        :param num: Number of phi to be sampled, defaults to 1.
        :type num: int, optional
        :type posterior: bool, optional
        :return: Primal representation.
        :rtype: Tensor[num, dim_input]
        """
        h = self.draw_h(num)
        return self.phi_map(h)

    @torch.no_grad()
    def draw_k(self, num: int = 1, posterior: bool = False) -> T:
        r"""
        Draws a dual representation k given its posterior distribution.

        :param posterior: Indicates whether phi has to be drawn from its posterior distribution or its conditional
            given the prior of h. Defaults to True.
        :param num: Number of k to be sampled, defaults to 1.
        :type num: int, optional
        :type posterior: bool, optional
        :return: Dual representation.
        :rtype: Tensor[num, num_idx]
        """
        h = self.draw_h(num)
        return self.k_map(h)



# coding=utf-8
from __future__ import annotations

import os
import wandb
import torch

from .Plotter import Plotter, Watcher
from ..utils import extend_docstring


@extend_docstring(Watcher)
class WandB(Plotter):
    def __init__(self, *args, **kwargs):
        super(WandB, self).__init__(*args, **kwargs)
        os.environ['WANDB_SILENT'] = "true"

        # INITIALIZE WANDB
        if self.verbose: print('Initializing Weights and Biases...', end=" ")
        if not os.path.exists(self.dir_plotter):
            os.makedirs(self.dir_plotter)
        self._wandb_run = wandb.init(name=self.expe_name,
                                     dir=self.dir_plotter,
                                     id=self.expe_id,
                                     project='Kerch')
        self._wandb_run.define_metric(name='objective_loss', summary='min')
        self._wandb_run.define_metric(name='training_error', summary='min')
        self._wandb_run.define_metric(name='validation_error', summary='min')
        self._wandb_run.define_metric(name='test_error', summary='min')

        # LOG THE HYPERPARAMETERS
        model_hparams = self.model.hparams_fixed
        opt_hparams = {} if self.opt is None else self.opt.hparams
        hparams = {**opt_hparams, **model_hparams}
        self._wandb_run.config.update(hparams)

        if self.verbose: print('Done')

    @property
    def _plotter_name(self) -> str | None:
        return None

    def finish(self) -> str:
        filepath = super(WandB, self).finish()
        self._wandb_run.log_model(path=filepath, name="final")
        self._wandb_run.finish(quiet=not self.verbose)
        return filepath

    def update(self,
               epoch: int,
               objective_loss: float,
               training_error: float | None = None,
               validation_error: float | None = None,
               test_error: float | None = None
               ) -> None:
        if epoch % self._num_epochs_save == 0:
            self.save_model(epoch)
        if epoch % self._num_epochs_params == 0:
            wandb_data = dict()
            for key, val in self.model.hparams_variable.items():
                wandb_data[key] = wandb.Image(val) if isinstance(val, torch.Tensor) else val
            self._wandb_run.log(data=wandb_data, step=epoch)
        if epoch % self._num_epochs_loss == 0:
            self._wandb_run.log(data=self.model.losses, step=epoch)
            self._wandb_run.log(data={'objective_loss': objective_loss,
                                      'training_error': training_error,
                                      'validation_error': validation_error,
                                      'test_error': test_error}, step=epoch)
        if epoch % self._num_epochs_plot == 0:
            self._wandb_run.log(data=self.model.watched_properties, step=epoch)



# coding=utf-8
from __future__ import annotations

import torch
from typing import Union

from ..utils import DEFAULT_KERNEL_TYPE
from kerch.utils import castf


@torch.no_grad()
def smoother(coefficients: torch.Tensor, observations: torch.Tensor,
             num: Union[int, str] = 'all') -> torch.Tensor:
    r"""
    Returns a weighted sum of the observations by the coefficients.

    .. math::
        \mathtt{out}_{i,j} = \frac{\sum_l\mathtt{weights}_{i,l} * \mathtt{observations}_{l,j}}{\sum_l\mathtt{weights}_{i,l}}.

    :param coefficients: coefficients used in the smoother.
    :type coefficients: torch.Tensor [num_points, num_observations]
    :param observations: observation corresponding to each weight dimension.
    :type observations: torch.Tensor [num_observations, dim_observations]
    :param num: Number of closest points to be used. Either an integer representing the number or
        the string ``'all'``. Defaults to ``'all'``.
    :type num: int or str, optional
    :return: Weighted observations
    :rtype: torch.Tensor [num_points, dim_observations]
    """

    # PRELIMINARIES
    coefficients = castf(coefficients)
    observations = castf(observations)

    num_points, num_coefficients = coefficients.shape
    num_observations = observations.shape[0]

    # DEFENSIVE
    assert num_coefficients == num_observations, \
        f'Smoother: Incorrect number of coefficients ({num_coefficients}), ' \
        f'compared to the number of points observations points({num_observations}).'

    if isinstance(num, str):
        assert num.lower() == 'all', \
            f"Only the string value of 'all' is allowed as value for the argument num ({num})." \
            f"Please use an integer is you prefer specifying a specific number."
        num = num_coefficients
    else:
        try:
            num = int(num)
        except ValueError:
            raise ValueError('The argument num is not an integer.')

    assert num > 0, \
        f"The argument num ({num}) must be a strictly positive integer (num > 0)."

    assert num <= num_coefficients, \
        f"The argument num ({num}) exceeds the number of observations ({num_observations})."

    # PRE-IMAGE
    if num == num_coefficients:
        kept_coeff = coefficients
    else:
        vals, indices = torch.topk(coefficients, k=num, dim=1, largest=True)
        kept_coeff = torch.zeros_like(coefficients)
        kept_coeff.scatter_(1, indices, vals)
    normalized_coeff = kept_coeff / torch.sum(kept_coeff, dim=1, keepdim=True)
    return normalized_coeff @ observations



@torch.no_grad()
def kernel_smoother(domain: torch.Tensor, observations: torch.Tensor, num: Union[int, str] = 'all', kernel_type: str=DEFAULT_KERNEL_TYPE, **kwargs) -> torch.Tensor:
    r"""
    Smoothens the function :math:`f(\mathtt{domain}_{i,:}) = \mathtt{observation}_{i,:}` with weights defined
    as a kernel on the domain.

    .. math::
        \mathtt{out}_{i,j} = \frac{\sum_l k\left(\mathtt{domain}_{i,:},\mathtt{domain}_{l,:}\right) * \mathtt{observations}_{l,j}}{\sum_l k\left(\mathtt{domain}_{i,:},\mathtt{domain}_{l,:}\right)}.

    The kernel is defined as in :py:func:`kerch.kernel.factory`.

    :param domain: domain corresponding to each observation.
    :type domain: torch.Tensor [num_observations, dim_domain]
    :param observations: observation corresponding to each domain entry.
    :type observations: torch.Tensor [num_observations, dim_observations]
    :param num: Number of closest points to be used. Either an integer representing the number or
        the string ``'all'``. Defaults to ``'all'``.
    :type num: int or str, optional
    :param kernel_type: Type of kernel chosen. For the possible choices, please refer to the `Factory Type` column of the
        :doc:`../kernel/index` documentation. Defaults to :py:data:`kerch.DEFAULT_KERNEL_TYPE`.
    :param \**kwargs: Arguments to be passed to the kernel constructor, such as `sample` or `sigma`. If an argument is
        passed that does not exist (e.g. `sigma` to a `linear` kernel), it will just be neglected. For the default
        values, please refer to the default values of the requested kernel.
    :type kernel_type: str, optional
    :type \**kwargs: dict, optional
    :return: Smoothed function :math:`f` according to kernel :math:`k`.
    :rtype: torch.Tensor [num_observations, dim_observations]
    """
    domain = castf(domain)
    observations = castf(observations)

    assert domain.shape[0] == observations.shape[0], f"Not the same number of domain {domain.shape[0]} and coefficients points {domain.shape[0]}."

    from ..kernel import factory

    k = factory(kernel_type=kernel_type, sample=domain, **kwargs)
    return smoother(coefficients=k.K, observations=observations, num=num)


# coding=utf-8
import argparse
import yaml

from . import version

def get_args() -> dict:
    parser = argparse.ArgumentParser(description=f"Kernel methods and deep machines using PyTorch.",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--version',
                        action='version',
                        version=version.run(),
                        help='prints the version and other information')
    parser.add_argument('action',
                        metavar='action',
                        action='store',
                        type=str,
                        choices=['expe'],
                        nargs=1,
                        help="action to perform ('expe').")
    parser.add_argument('-f', '--filename',
                        action='store',
                        type=str,
                        help='path to the yaml file containing the experiment arguments',
                        required=False)
    parser.add_argument('-v', '--verbose',
                        action='store_true',
                        help='prints the progress of the script')
    args = parser.parse_args()
    return vars(args)

def get_kwargs(filename):
    with open(filename, "r") as file:
        return yaml.safe_load(file)


# coding=utf-8
from .single_view.Level import Level
def factory(level_type='kpca', **kwargs) -> Level:
    r"""
    Creates a kernel based on the specified name with the specified arguments. This is the same as
    calling `kerch.kernel.name(*args, **kwargs)` (if `name` is not a string here). This allows for the creation of kernel where
    the name of kernel is passed as a string.

    :param level_type: Type of level chosen. For the possible choices, please refer to the (non-abstract) classes
        herebelow., defaults to `kpca`
    :param \**kwargs: Arguments to be passed to the level constructor. If an argument is
        passed that does not exist, it will just be neglected. For the default
        values, please refer to the default values of the requested level.
    :type level_type: str, optional
    :type \**kwargs: dict, optional
    :return: An instantiation of the specified level.

    .. warning::
        Only supports single view levels.
    """

    if level_type.lower() not in ["kpca", "lssvm", "ridge"]:
        raise NameError(f"Invalid level type, currently only KPCA, LSSVM and Ridge supported.")

    def case_insensitive_getattr(obj, attr):
        for a in dir(obj):
            if a.lower() == attr.lower():
                return getattr(obj, a)
        return None

    import kerch.level.single_view
    level = case_insensitive_getattr(kerch.level.single_view, level_type)
    if level is None:
        raise NameError("Invalid level type.")

    return level(**kwargs)



# coding=utf-8
from __future__ import annotations

import torch
from torch import Tensor as T
from typing import Optional, Iterator
from abc import ABCMeta, abstractmethod

from ._Level import _Level
from .. import utils


class _PPCA(_Level, metaclass=ABCMeta):
    @utils.kwargs_decorator({'use_mean': False,
                             'feature_noise': None})
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.use_mean = kwargs["use_mean"]
        self.feature_noise = kwargs["feature_noise"]
        self._vals = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE),
                                        requires_grad=False)
        self._parameter_related_cache = [*self._parameter_related_cache,
                                         "_B_primal", "_B_dual", "_Inv_primal", "_Inv_dual"]

    @property
    def use_mean(self) -> bool:
        return self._use_mean

    @use_mean.setter
    def use_mean(self, val: bool) -> None:
        self._use_mean = val

    @property
    def vals(self) -> T:
        r"""
        Eigenvalues of the model. The model has to be fitted for these values to exist.
        """
        return self._vals

    @property
    def sqrt_vals(self) -> T:
        return self._get(key="sqrt_vals", fun=lambda: torch.sqrt(self.vals))

    @vals.setter
    def vals(self, val):
        val = utils.castf(val, tensor=False, dev=self._vals.device)
        self._vals.data = val

    def total_variance(self, as_tensor=False, normalize=True, representation=None) -> float | T:
        r"""
        Total variance contained in the feature map. In primal formulation,
        this is given by :math:`\DeclareMathOperator{\tr}{tr}\tr(C)`, where :math:`C = \sum\phi(x)\phi(x)^\top` is
        the covariance matrix on the sample. In dual, this is given by :math:`\DeclareMathOperator{\tr}{tr}\tr(K)`,
        where :math:`K_{ij} = k(x_i,x_j)` is the kernel matrix on the sample.

        :param as_tensor: Indicated whether the variance has to be returned as a float or a torch.Tensor., defaults
            to ``False``
        :type as_tensor: bool, optional

        .. warning::
            For this value to strictly be interpreted as a variance, the corresponding kernel (or feature map)
            has to be normalized. In that case however, the total variance will amount to the dimension of the feature
            map in primal and the number of datapoints in dual.
        """
        representation = utils.check_representation(representation=representation, default=self._representation)
        level_key = "KPCA_total_variance_default_representation" if representation == self._representation \
            else "KPCA_total_variance_other_representation"
        if representation == 'primal':
            var = self._get("total_variance_primal", level_key=level_key, fun=lambda: torch.trace(self.C))
        else:
            var = self._get("total_variance_dual", level_key=level_key, fun=lambda: torch.trace(self.K))
        if normalize:
            var /= self.num_idx
        if as_tensor:
            return var
        return var.detach().cpu().numpy()

    def model_variance(self, as_tensor=False, normalize=True) -> float | T:
        r"""
        Total variance learnt by the model given by the sum of the eigenvalues.

        :param as_tensor: Indicated whether the variance has to be returned as a float or a torch.Tensor., defaults
            to ``False``
        :type as_tensor: bool, optional

        .. warning::
            For this value to strictly be interpreted as a variance, the corresponding kernel (or feature map)
            has to be normalized. We refer to the remark of ``total_variance``.
        """
        var = torch.sum(self.vals)
        if normalize:
            var /= self.num_idx
        if as_tensor:
            return var
        return var.detach().cpu().numpy()

    def _reset_dual(self) -> None:
        super(_PPCA, self)._reset_dual()
        self._remove_from_cache(["total_variance_primal", "total_variance_dual"])

    def relative_variance(self, as_tensor=False) -> float | T:
        r"""
        Relative variance learnt by the model given by ```model_variance``/``total_variance``.
        This number is always comprised between 0 and 1 and avoids any considerations on normalization.

        :param as_tensor: Indicated whether the variance has to be returned as a float or a torch.Tensor., defaults
            to ``False``
        :type as_tensor: bool, optional
        """
        var = self.model_variance(as_tensor=as_tensor, normalize=False) / \
              self.total_variance(as_tensor=as_tensor, normalize=False)
        return var

    @property
    def feature_noise(self) -> float:
        r"""
        Sigma value, acts as a regularization parameter.
        """
        return self._feature_noise

    @feature_noise.setter
    def feature_noise(self, val: Optional[float]) -> None:
        if val is None:
            self._feature_noise = None
        else:
            self._feature_noise = float(val)

    @property
    def mu(self) -> torch.nn.Parameter:
        return self._mu

    @mu.setter
    def mu(self, val: torch.Tensor) -> None:
        self._mu.data = utils.castf(val, dev=self.sample.device, tensor=True)

    ########################################################################################################
    @property
    @torch.no_grad()
    def _B_primal(self) -> T:
        def compute() -> T:
            return torch.cholesky(self.weight @ self.weight.T + self.feature_noise ** 2 * self._I_primal)

        return self._get(key="_B_primal", level_key="PPCA_B_primal", fun=compute)

    @_B_primal.setter
    def _B_primal(self, val: T) -> None:
        self._get(key="_B_primal", level_key="PPCA_B_primal", force=True, fun=lambda: val)

    @property
    @torch.no_grad()
    def _B_dual(self) -> T:
        def compute() -> T:
            return torch.cholesky(self.hidden.T @ self.hidden + self.feature_noise ** 2 * torch.inv(self.K))

        return self._get(key="_B_dual", level_key="PPCA_B_dual", fun=compute)

    @_B_dual.setter
    def _B_dual(self, val: T) -> None:
        self._get(key="_B_dual", level_key="PPCA_B_dual", force=True, fun=lambda: val)

    @property
    @torch.no_grad()
    def _Inv_primal(self) -> T:
        def compute() -> T:
            return torch.inv(self.weight.T @ self.weight + self.feature_noise ** 2 * self._I_primal)

        return self._get(key="_Inv_primal", level_key="PPCA_Inv_primal", fun=compute)

    @_Inv_primal.setter
    @torch.no_grad()
    def _Inv_primal(self, val: T) -> None:
        self._get(key="_Inv_primal", level_key="PPCA_Inv_primal", fun=lambda: val, overwrite=True)

    @property
    @torch.no_grad()
    def _Inv_dual(self) -> T:
        def compute() -> T:
            return torch.inv(self.hidden.T @ self.K @ self.hidden + self.feature_noise ** 2 * self._I_dual)

        return self._get(key="_Inv_dual", level_key="PPCA_Inv_dual", fun=compute)

    @_Inv_dual.setter
    @torch.no_grad()
    def _Inv_dual(self, val: T) -> None:
        self._get(key="_Inv_dual", level_key="PPCA_Inv_dual", fun=lambda: val, overwrite=True)

    ########################################################################################################################

    @torch.no_grad()
    def h_map(self, phi: Optional[T] = None, k: Optional[T] = None) -> T:
        r"""
        Draws a `h` given the maximum a posteriori of the distribution. By choosing the input, you either
        choose a primal or dual representation.

        :param phi: Primal representation.
        :param k: Dual representation.
        :type phi: Tensor[N, dim_input], optional
        :type k: Tensor[N, num_idx], optional
        :return: MAP of h given phi or k.
        :rtype: Tensor[N, dim_output]
        """

        if phi is not None and k is None:
            return phi @ self.weight.T @ self._Inv_primal
        if phi is None and k is not None:
            return k @ self.hidden.T @ self._Inv_dual
        else:
            raise AttributeError("One and only one attribute phi or k has to be specified.")

    @torch.no_grad()
    def phi_map(self, h: T) -> T:
        r"""
        Maximum a posteriori of phi given h.

        :param h: Latent representation.
        :type h: Tensor[N, dim_output]
        :return: MAP of phi given h.
        :rtype: Tensor[N, dim_input]
        """
        if self.use_mean:
            return h @ self.weight.T + self.mu
        return h @ self.weight.T

    @torch.no_grad()
    def k_map(self, h: T) -> T:
        r"""
        Maximum a posteriori of k given h.

        :param h: Latent representation.
        :type h: Tensor[N, dim_output]
        :return: MAP of k given h.
        :rtype: Tensor[N, num_idx]
        """
        if self.use_mean:
            raise NotImplementedError
        return h @ self.hidden.T @ self.K

    @torch.no_grad()
    def draw_h(self, num: int = 1) -> T:
        r"""
        Draws a h given its prior distribution.

        :param num: Number of h to be sampled, defaults to 1.
        :type num: int, optional
        :return: Latent representation.
        :rtype: Tensor[num, dim_output]
        """
        return torch.randn((num, self.dim_output), device=self.sample.device, dtype=utils.FTYPE)

    @torch.no_grad()
    def draw_phi(self, num: int = 1, posterior: bool = True) -> T:
        r"""
        Draws a primal representation phi given its posterior distribution.

        :param posterior: Indicates whether phi has to be drawn from its posterior distribution or its conditional
            given the prior of h. Defaults to True.
        :param num: Number of phi to be sampled, defaults to 1.
        :type num: int, optional
        :type posterior: bool, optional
        :return: Primal representation.
        :rtype: Tensor[num, dim_input]
        """
        if posterior:
            u = torch.randn((num, self.dim_feature), dtype=utils.FTYPE, device=self.sample.device)
            if self.use_mean:
                return u @ self._B_primal + self.mu
            return u @ self._B_primal
        h = self.draw_h(num)
        return self.phi_map(h)

    @torch.no_grad()
    def draw_k(self, num: int = 1, posterior: bool = False) -> T:
        r"""
        Draws a dual representation k given its posterior distribution.

        :param posterior: Indicates whether phi has to be drawn from its posterior distribution or its conditional
            given the prior of h. Defaults to True.
        :param num: Number of k to be sampled, defaults to 1.
        :type num: int, optional
        :type posterior: bool, optional
        :return: Dual representation.
        :rtype: Tensor[num, num_idx]
        """
        if posterior:
            u = torch.randn((num, self.num_idx), dtype=utils.FTYPE, device=self.sample.device)
            if self.use_mean:
                raise NotImplementedError
            return u @ self._B_dual
        h = self.draw_h(num)
        return self.k_map(h)

    ################################################################################################################

    def loss(self, representation=None) -> T:
        return torch.tensor(0., dtype=utils.FTYPE)

    def _stiefel_parameters(self, recurse=True) -> Iterator[torch.nn.Parameter]:
        super(_PPCA, self)._euclidean_parameters(recurse)
        if self._representation == 'primal':
            if self._primal_param_exists:
                yield self._primal_param
        else:
            if self._dual_param_exists:
                yield self._dual_param



from .Model import Model

class LSSVM(Model):
    def __init__(self, *args, **kwargs):
        super(LSSVM, self).__init__(*args, **kwargs)
        kwargs['level_type'] = 'lssvm'
        self._append_level(*args, **kwargs)

    def __str__(self):
        return "[Model] " + self._first_level.__repr__()


"""
Allows for a sample set manager, with stochastic support.
"""

import torch

from abc import ABCMeta, abstractmethod
from torch import Tensor
from typing import Union

from . import utils
from ._cache import _Cache
from ._logger import _Logger
from ._stochastic import _Stochastic
from .projections import ProjectionTree

@utils.extend_docstring(_Stochastic)
class _Sample(_Stochastic,  # manager stochastic indices
              _Cache,  # creates a transportable cache (e.g. for GPU)
              _Logger,  # allows logging actions and errors
              metaclass=ABCMeta):
    r"""
    :param sample: Sample points used to compute the kernel matrix. When an out-of-sample computation is asked, it will
        be given relative to these samples., defaults to `None`
    :param sample_trainable: `True` if the gradients of the sample points are to be computed. If so, a graph is
        computed and the sample can be updated. `False` just leads to a static computation., defaults to `False`
    :param num_sample: Number of sample points. This parameter is neglected if `sample` is not `None` and overwritten by
        the number of points contained in sample., defaults to 1
    :param dim_input: Dimension of each sample point. This parameter is neglected if `sample` is not `None` and
        overwritten by the dimension of the sample points., defaults to 1

    :type sample: Tensor(num_sample, dim_input), optional
    :type sample_trainable: bool, optional
    :type num_sample: int, optional
    :type dim_input: int, optional

    :param idx_sample: Initializes the indices of the samples to be updated. All indices are considered if both
        `idx_stochastic` and `prop_stochastic` are `None`., defaults to `None`
    :param prop_sample: Instead of giving indices, specifying a proportion of the original sample set is also
        possible. The indices will be uniformly randomly chosen without replacement. The value must be chosen
        such that :math:`0 <` `prop_stochastic` :math:`\leq 1`. All indices are considered if both `idx_stochastic` and
        `prop_stochastic` are `None`., defaults to `None`.

    :type idx_sample: int[], optional
    :type idx_sample: float, optional

    :param sample_projection: TODO
    :type sample_projection: List[str]
    """

    @abstractmethod
    @utils.kwargs_decorator({
        "sample": None,
        "sample_trainable": False,
        "num_sample": None,
        "dim_input": None,
        "idx_sample": None,
        "prop_sample": None,
        "sample_projections": []})
    def __init__(self, **kwargs):
        super(_Sample, self).__init__(**kwargs)

        sample = kwargs["sample"]
        if sample is not None:
            sample = utils.castf(sample)
            self._num_total, self._dim_input = sample.shape
        else:
            self._dim_input = kwargs["dim_input"]
            self._num_total = kwargs["num_sample"]

        self._sample = torch.nn.Parameter(torch.empty((0, 0)))
        self._sample_trainable = kwargs["sample_trainable"]

        self.init_sample(sample, idx_sample=kwargs["idx_sample"], prop_sample=kwargs["prop_sample"])
        self._default_sample_projections = kwargs["sample_projections"]

    @property
    def dim_input(self) -> int:
        r"""
        Dimension of each datapoint.
        """
        return self._dim_input

    @dim_input.setter
    def dim_input(self, val: int):
        assert self._dim_input is None, "Cannot set the dimension of the sample points after initialization if the " \
                                        "sample dataset. Use init_sample() instead."
        self._dim_input = val
        if self._num_total is not None:
            self.init_sample()

    @property
    def _empty_sample(self) -> bool:
        r"""
        Boolean specifying if the sample is empty or not.
        """
        return self._sample.nelement() == 0

    @property
    def num_sample(self) -> int:
        r"""
        Number of datapoints in the sample set.
        """
        if self._empty_sample:
            return 0
        return self._num_total

    @num_sample.setter
    def num_sample(self, val: int):
        assert self._num_total is None, "Cannot set the number of sample points after initialization if the " \
                                        "sample dataset. Use init_sample() instead."
        self._num_total = val
        if self._dim_input is not None:
            self.init_sample()

    @property
    def sample(self) -> torch.nn.Parameter:
        r"""
        Raw sample before any projectionation.
        """
        return self._sample

    @sample.setter
    def sample(self, val):
        assert val is not None, "Cannot assign the sample to None. Please use init_sample() if you want to " \
                                "re-initialize it."
        self.init_sample(val)

    @property
    def sample_trainable(self) -> bool:
        r"""
        Boolean if the sample dataset can be trained.
        """
        return self._sample_trainable

    @sample_trainable.setter
    def sample_trainable(self, val: bool):
        self._sample_trainable = val
        self._sample.requires_grad = self._sample_trainable

    @property
    def current_sample(self) -> Tensor:
        r"""
        Returns the sample that is currently used in the computations and for the normalizing and centering statistics
        if relevant.
        """
        return self.sample_projections.projected_sample

    @property
    def current_sample_unprojectioned(self) -> Tensor:
        return self._sample[self.idx, :]

    def init_sample(self, sample=None, idx_sample=None, prop_sample=None):
        r"""
        Initializes the sample set (and the stochastic indices).

        :param sample: Sample points used for the various computations. When an out-of-sample computation is asked, it
            will be given relative to these samples. In case of overwriting a current sample, `num_sample` and
            `dim_input` are also overwritten. If `None` is specified, the sample dataset will be initialized according
            to `num_sample` and `dim_input` specified during the construction. If a previous sample set has been used,
            it will keep the same dimension by consequence. A last case occurs when `sample` is of the class
            `torch.nn.Parameter`: the sample will then use those values, and they can thus be shared with the module
            calling this method., defaults to `None`
        :type sample: Tensor, optional
        :param idx_sample: Initializes the indices of the samples to be updated. All indices are considered if both
            `idx_sample` and `prop_sample` are `None`., defaults to `None`
        :type idx_sample: int[], optional
        :param prop_sample: Instead of giving indices, specifying a proportion of the original sample set is also
            possible. The indices will be uniformly randomly chosen without replacement. The value must be chosen
            such that :math:`0 <` `prop_sample` :math:`\leq 1`. All indices are considered if both `idx_sample` and
            `prop_sample` are `None`., defaults to `None`.
        """
        sample = utils.castf(sample)

        if sample is None:
            self._log.debug("Initializing new sample with the sample dimensions.")
            if self._num_total is None or self.dim_input is None:
                self._log.info(
                    'The sample cannot be initialized because no sample dataset has been provided nor the '
                    'sample dimensions have been initialized yet.')
                return
            self._sample = torch.nn.Parameter(
                torch.nn.init.orthogonal_(torch.empty((self._num_total, self._dim_input),
                                                      dtype=utils.FTYPE,
                                                      device=self._sample.device)),
                requires_grad=self._sample_trainable)
        elif isinstance(sample, torch.nn.Parameter):
            self._log.debug("Using existing sample defined as an external parameter.")
            self._num_total, self._dim_input = sample.shape
            self._sample_trainable = sample.requires_grad
            self._sample = sample
        else:
            self._log.debug("Assigning new sample points based on the given values.")
            self._num_total, self._dim_input = sample.shape
            self._sample = torch.nn.Parameter(sample.data,
                                              requires_grad=self._sample_trainable)

        for sample_module in self.children():
            if isinstance(sample_module, _Sample):
                sample_module.init_sample(self.current_sample)

        self.stochastic(idx=idx_sample, prop=prop_sample)

    def update_sample(self, sample_values, idx_sample=None):
        r"""
        Updates the sample set. In contradiction to `init_samples`, this only updates the values of the sample and sets
        the gradients of the updated values to zero if relevant.

        :param sample_values: Values given to the updated samples.
        :param idx_sample: Indices of the samples to be updated. All indices are considered if `None`., defaults to
            `None`
        :type sample_values: Tensor
        :type idx_sample: int[], optional
        """

        if self._empty_sample:
            self._log.warning('Cannot update the sample values of a None sample dataset.')
            return

        sample_values = utils.castf(sample_values, dev=self._sample.device)
        self._reset_cache()

        # use all indices if unspecified
        if idx_sample is None:
            idx_sample = self._all_idx

        # check consistency of indices
        assert len(idx_sample) == sample_values.shape[0], f"Number of sample values ({sample_values.shape[0]}) and " \
                                                          f"corresponding indices ({len(idx_sample)}) are not " \
                                                          f"consistent."
        # update the values
        self._sample.data[idx_sample, :] = sample_values.data

        # zeroing relevant gradient if relevant
        if self._sample_trainable:
            self._sample.grad.data[idx_sample, :].zero_()

    def _euclidean_parameters(self, recurse=True):
        if not self._empty_sample:
            yield self._sample
        yield from super(_Sample, self)._euclidean_parameters(recurse)

    @property
    def sample_projections(self) -> ProjectionTree:
        r"""
        Default projections used by the sample.
        """
        return self._sample_projections

    @property
    def _sample_projections(self) -> ProjectionTree:
        def fun():
            return ProjectionTree(explicit=True,
                                  sample=self._sample[self.idx, :],
                                  default_projections=self._default_sample_projections,
                                  cache_level=self._cache_level)
        return self._get("sample_projections", "oblivious", fun)

    def project_sample(self, data) -> Union[Tensor, None]:
        r"""
        Apply to value the same projections as on the sample.
        """
        if data is None:
            return None
        return self.sample_projections.apply(data)

    def projection_sample_revert(self, data) -> Tensor:
        r"""
        Get back the original value from a projected value, by using the same projections as the sample,
        but in reverse. This is not always feasible, depending on the projections used (normalizations are
        typically not invertible as they are projections which are not bijective).
        """
        return self.sample_projections.revert(data)



# coding=utf-8
from __future__ import annotations
from typing import Union, Tuple
from abc import ABCMeta, abstractmethod

import torch
from ..feature import Cache
from ..utils.errors import BijectionError, ImplicitError
from ..utils import extend_docstring


class Transform(Cache, metaclass=ABCMeta):
    def __init__(self, explicit: bool, name: str, default_path: bool = False, **kwargs):
        logger_name = 'Explicit' if explicit else 'Implicit'
        super(Transform, self).__init__(**{'logger_name': logger_name, **kwargs})
        self._parent = None
        self._offspring: dict = {}
        self._name: str = name
        self._explicit: bool = explicit
        self._default: bool = False
        self._default_path: bool = default_path
        self._set_cache_levels()

    def _set_cache_levels(self):
        # CACHE LEVELS
        if self._default:
            self._sample_data_level = "transform_sample_data_default"
            self._sample_statistics_level = "transform_sample_statistics_default"
            self._oos_data_level = "transform_oos_data_default"
            self._oos_statistics_level = "transform_oos_statistics_default"
        else:
            self._sample_data_level = "transform_sample_data_nondefault"
            self._sample_statistics_level = "transform_sample_statistics_nondefault"
            self._oos_data_level = "transform_oos_data_nondefault"
            self._oos_statistics_level = "transform_oos_statistics_nondefault"

    def __str__(self):
        if self.default:
            return self._name + " (default)"
        return self._name

    @staticmethod
    def _get_names(x=None, y=None) -> Tuple[str, str]:
        if x is not None:
            x_name = id(x)
        else:
            x_name = 0
        if y is not None:
            y_name = id(y)
        else:
            y_name = 0
        return str(x_name), str(y_name)

    @property
    def parent(self) -> Union[Transform, None]:
        return self._parent

    @property
    def default(self) -> bool:
        return self._default

    @default.setter
    def default(self, val: bool):
        self._default = val
        self._set_cache_levels()


    @property
    def explicit(self) -> bool:
        return self._explicit

    @property
    def offspring(self) -> dict:
        return self._offspring

    def add_offspring(self, val: Transform) -> None:
        # add offspring to parent
        if type(val) in self.offspring:
            self._logger.debug(f"Overwriting offspring of type {type(val)} in parent {type(self._parent)}.")
        self._offspring[type(val)] = val

        # add parent to offspring
        if val.parent is not None:
            self._logger.debug(f"Overwriting parent of offspring {type(self)}.")
        val._parent = self

    def add_parent(self, val: Transform) -> None:
        val.add_offspring(self)

    @property
    def is_leaf(self) -> bool:
        return not bool(self.offspring)

    @property
    def is_root(self) -> bool:
        return self.parent is None

    # DATA
    @abstractmethod
    def _explicit_statistics(self, sample):
        pass

    def _implicit_statistics(self, sample, x=None):
        raise BijectionError

    @abstractmethod
    def _explicit_sample(self):
        pass

    def _implicit_sample(self):
        raise ImplicitError

    @property
    def sample(self) -> torch.Tensor:
        return self._get(self._name + " data_sample", level_key=self._sample_data_level,
                         fun=self._explicit_sample if self.explicit else self._implicit_sample)

    def statistics_sample(self, sample=None) -> torch.Tensor:
        def fun(sample):
            if sample is None:
                sample = self.parent.sample
            if self.explicit:
                return self._explicit_statistics(sample=sample)
            else:
                return self._implicit_statistics(sample=sample)

        return self._get(self._name + " statistics_sample", level_key=self._sample_data_level, fun=lambda: fun(sample))

    # OOS
    @abstractmethod
    def _explicit_statistics_oos(self, oos, x=None):
        pass

    def _implicit_statistics_oos(self, oos, x=None):
        raise BijectionError

    @abstractmethod
    def _explicit_oos(self, x=None):
        pass

    def _implicit_oos(self, x=None, y=None):
        raise BijectionError

    def statistics_oos(self, x=None, y=None, oos=None) -> Union[torch.Tensor, (torch.Tensor, torch.Tensor)]:
        x_name, y_name = Transform._get_names(x, y)

        if self.explicit:
            if x is None:
                return self.statistics_sample()
            else:
                def fun(x, oos):
                    if oos is None:
                        oos = self.parent.oos(x=x)
                    return self._explicit_statistics_oos(x=x, oos=oos)

                return self._get(self._name + " statistics_oos_" + x_name, level_key=self._oos_statistics_level,
                                 fun=lambda: fun(x, oos), force=True)
        else:
            def fun(x, oos):
                if oos is None:
                    oos = self.parent.oos(x=x)
                return self._implicit_statistics_oos(x=x, oos=oos)

            if x is None:
                x_stat = self.statistics_sample()
            else:
                x_stat = self._get(self._name + " statistics_oos_" + x_name, level_key=self._oos_statistics_level,
                                   fun=lambda: fun(x, oos), force=True)
            if y is None:
                y_stat = self.statistics_sample()
            else:
                y_stat = self._get(self._name + " statistics_oos_" + y_name, level_key=self._oos_statistics_level,
                                   fun=lambda: fun(y, oos), force=True)
            return (x_stat, y_stat)

    def oos(self, x=None, y=None) -> torch.Tensor:
        x_name, y_name = Transform._get_names(x, y)

        if self.explicit:
            if x is None:
                return self._explicit_sample()
            else:
                return self._get(self._name + " data_oos_" + x_name, level_key=self._oos_data_level,
                                 fun=lambda: self._explicit_oos(x=x))
        else:
            if x is None and y is None:
                return self._implicit_sample()
            elif x is None:  # y is relevant
                return self._get(self._name + ' data_oos_' + y_name, level_key=self._oos_data_level,
                                 fun=lambda: self._implicit_oos(x=y)).T
            elif y is None:  # x is relevant
                return self._get(self._name + ' data_oos_' + x_name, level_key=self._oos_data_level,
                                 fun=lambda: self._implicit_oos(x=x))
            else:
                return self._get(self._name + ' data_oos_' + x_name + '_' + y_name, level_key=self._oos_data_level,
                                 fun=lambda: self._implicit_oos(x=x, y=y))

    def _revert(self, oos):
        if self.explicit:
            return self._revert_explicit(oos)
        else:
            return self._revert_implicit(oos)

    def _revert_explicit(self, oos):
        raise BijectionError

    def _revert_implicit(self, oos):
        raise ImplicitError

    def _implicit_diag(self, x=None):
        return torch.diag(self.oos(x, x))[:, None]

    @property
    @extend_docstring(Cache.cache_level)
    def cache_level(self) -> str:
        switcher = Cache._cache_level_switcher
        inv_switcher = {switcher[k]: k for k in switcher}
        return inv_switcher[self._cache_level]

    @cache_level.setter
    def cache_level(self, val: Union[str, int]):
        val = self._get_level(val)
        self._cache_level = val
        try:
            for o in self._offspring.values():
                o.cache_level = val
        except AttributeError:
            pass

    @extend_docstring(Cache.print_cache)
    def print_cache(self, private: bool = False) -> None:
        super(Transform, self).print_cache(private)
        try:
            for o in self._offspring.values():
                o.print_cache(private)
        except AttributeError:
            pass
    @extend_docstring(Cache._clean_cache)
    def _clean_cache(self, max_level: Union[str, int, None] = None):
        super(Transform, self)._clean_cache(max_level)
        try:
            for o in self._offspring.values():
                o._clean_cache(max_level)
        except AttributeError:
            pass


# coding=utf-8
from abc import ABCMeta, abstractmethod
from ..feature.logger import Logger
from .decorators import extend_docstring

class KerchError(Exception, metaclass=ABCMeta):
    r"""
    :param cls: Optional class throwing the exception. This is helpful to add a context to the error message.
    :param message: Optional error message.

    :type cls: Instance of :class:`kerch.feature.Logger`
    :type message: str
    """
    @abstractmethod
    def __init__(self, cls=None, message=""):
        msg = message
        if msg == "" and hasattr(self, 'message'):
            msg = self.message

        if isinstance(cls, Logger):
            msg = "[" + cls.__class__.__name__ + "] " + msg

        super(KerchError, self).__init__(msg)


@extend_docstring(KerchError)
class ImplicitError(KerchError):
    r"""
    Error thrown when an implicit representation is requested and mathematically not available.
    """
    def __init__(self, *args, **kwargs):
        self.message = "Implicit representation not available."
        super(ImplicitError, self).__init__(*args, **kwargs)


@extend_docstring(KerchError)
class ExplicitError(KerchError):
    r"""
    Error thrown when an explicit representation is requested and mathematically not available.
    """
    def __init__(self, *args, **kwargs):
        self.message = "Explicit representation not available.\n" \
                       "[Example 1]: The explicit representation does not exist as it lies in an infinite " \
                       "dimensional Hilbert space.\n" \
                       "[Example 2]: Only the inner product (implicit representation) is known, but not the " \
                       "original vectors."
        super(ExplicitError, self).__init__(*args, **kwargs)


@extend_docstring(KerchError)
class RepresentationError(KerchError):
    r"""
    Error thrown when the requested representation is invalid.
    """
    def __init__(self, *args, **kwargs):
        self.message = "Unrecognized or unspecified representation (must be primal or dual)."
        super(RepresentationError, self).__init__(*args, **kwargs)


@extend_docstring(KerchError)
class BijectionError(KerchError):
    r"""
    Error thrown when an inverse is requested from a non-bijective function.
    """
    def __init__(self, *args, **kwargs):
        self.message = "Mathematically undefined operation. A transform is not bijective, thus non invertible."
        super(BijectionError, self).__init__(*args, **kwargs)


@extend_docstring(KerchError)
class NotInitializedError(KerchError):
    r"""
    Error thrown when a model is asked to perform some tasks, but is not yet fully initialized.
    """
    def __init__(self, *args, **kwargs):
        self.message = "The model has not been initialized yet."
        super(NotInitializedError, self).__init__(*args, **kwargs)


@extend_docstring(KerchError)
class MultiViewError(KerchError):
    r"""
    Error thrown when some operations are requested, but not defined in a multi-view context.
    """
    def __init__(self, *args, **kwargs):
        self.message = "Primal operations are not defined a multi-view context. You must ask them for the different " \
                       "known separately, if it exists for it."
        super(MultiViewError, self).__init__(*args, **kwargs)



# coding=utf-8
"""
This abstract class has only one purpose, add a self._log property meant to log various information across the toolbox.
Doing it this way allows to get the name of the class instantiated and print more precise messages.

Author: HENRI DE PLAEN
Date: June 2022
"""
from __future__ import annotations

import logging
from abc import ABCMeta
import sys

from .. import _GLOBALS


class Logger(object, metaclass=ABCMeta):
    r"""
    :param logging_level: Logging level for this specific instance.
        If the value is ``None``, the current default kerch global log level will be used.
        Defaults to ``None`` (default kerch logging level).
        We refer to the :doc:`/features/logger` documentation for further information.
    :type logging_level: int, optional
    """

    if hasattr(sys, 'gettrace') and sys.gettrace() is not None: # debugger is active
        _LOGGING_FORMAT = "Kerch %(levelname)s [%(name)s]: %(message)s [%(pathname)s:%(lineno)d]"
    else:
        _LOGGING_FORMAT = "Kerch %(levelname)s [%(name)s]: %(message)s"

    _kerch_format = logging.Formatter(_LOGGING_FORMAT)
    _kerch_handler = logging.StreamHandler()
    _kerch_handler.setFormatter(_kerch_format)

    def __init__(self, *args, **kwargs):
        self._name = kwargs.pop('logger_name', None)
        class_name = self.__class__.__name__
        if self._name is not None and class_name != "Logger":
            log_name = self._name + ' ' + class_name
        elif self._name is not None:
            log_name = self._name
        else:
            log_name = class_name

        self._logger_internal: logging.Logger = logging.getLogger(name=log_name)
        self._logger_internal.addHandler(Logger._kerch_handler)
        self.logging_level = kwargs.pop('logging_level', None)

        self._logger.debug('Instantiating')

    @property
    def _logger(self) -> logging.Logger:
        r"""
        Logger of the instance.

        Usage:

        .. exec_code::

            # --- hide: start ---
            import sys
            import logging
            logging.basicConfig(stream=sys.stdout)
            # --- hide: stop ---

            import kerch
            import logging

            class MyClass(kerch.feature.Logger):
                def __init__(self, *args, **kwargs):
                    super(MyClass, self).__init__(*args, **kwargs)
                    self._logger.info('Instantiation done information.')
                    self._logger.warn('Instantiation done warning.')

            print('First class with default logging level:')
            my_class1 = MyClass()

            print('\nSecond instance with logging.INFO logging level:')
            my_class2 = MyClass(logging_level=logging.INFO)

        """
        return self._logger_internal

    @property
    def logging_level(self) -> int:
        r"""
        Logging level of this specific instance.
        If the value is ``None``, the current default kerch global log Level will be used.
        Defaults to ``None`` (default global kerch level).
        We refer to the :doc:`/features/logger` documentation for further information.
        """
        return self._logger_internal.level

    @logging_level.setter
    def logging_level(self, level: int | None):
        if level is None:
            level = _GLOBALS["LOG_LEVEL"]
        self._logger_internal.setLevel(level)


_GLOBAL_LOGGER = Logger(logger_name="global")


def set_logging_level(level: int):
    r"""
    Changes the default logging level of the kerch package.

    :param level: Default kerch logging value
    :type level: int

    Usage:

    .. code-block:: python

        import kerch
        import logging

        kerch.set_logging_level(logging.DEBUG)


    .. warning::
        Changing the default logging value does not affect the already instantiated objects. We advise to set those values in
        the beginning of the code.


    """
    _GLOBALS["LOG_LEVEL"] = level
    _GLOBAL_LOGGER.logging_level = level


def get_logging_level() -> int:
    r"""
    Returns the default logging level of the kerch package.

    Usage:

    .. exec_code::

        import kerch
        import logging

        default_level = kerch.get_logging_level()
        default_level = logging.getLevelName(default_level)
        print(default_level)

    """
    return _GLOBALS["LOG_LEVEL"]



# coding=utf-8
from typing import Union

import torch
from torch import Tensor as T


def eye_like(m: T) -> T:
    r"""
    Creates an identity matrix of the same size, data type and on the same device as the provided one.
    :param m: Provided matrix to create an identity matrix from.
    :return: Identity matrix of the same size, data type and on the same device.

    :type m: torch.Tensor
    :rtype: torch.Tensor
    """
    return torch.eye(*m.size(), out=torch.empty_like(m))


def ones_like(m: T) -> T:
    r"""
    Creates a matrix full of ones of the same size, data type and on the same device as the provided one.
    :param m: Provided matrix to create the new matrix from.
    :return: Matrix full of ones matrix of the same size, data type and on the same device.

    :type m: torch.Tensor
    :rtype: torch.Tensor
    """
    return torch.ones(*m.size(), out=torch.empty_like(m))


def equal(val1: Union[T, None], val2: Union[T, None]) -> bool:
    r"""
    Verifies whether the two provided tensors are identical in content. Can also provide `None` values which will only
    be identical to other `None` values.

    :param val1: First tensor to compare.
    :param val2: Second tensor to compare.
    :return: `True` if the two provided tensors are identical, `False` otherwise.

    :type val1: Union[T, None]
    :type val2: Union[T, None]
    :rtype: bool
    """
    if isinstance(val1, T) and isinstance(val2, T):
        return torch.equal(val1, val2)
    elif isinstance(val1, str) and isinstance(val2, str):
        return val1 == val2
    elif val1 is None and val2 is None:
        return True
    else:
        return False



# coding=utf-8
import lazy_loader
from .. import (__version__,
                __author__,
                __status__,
                __credits__,
                __date__,
                __license__,
                gpu_available)
torch = lazy_loader.load('torch')

def run():
    newline = '\n'
    if gpu_available():
        gpu = f"with GPU acceleration available (CUDA version: {torch.version.cuda}."
    else:
        gpu = f"with no GPU acceleration available."
    return (f"Kerch version: {__version__} {__status__} ({__date__}). {newline}"
            f"Author: {__author__}. Copyright: {__credits__}. License: {__license__}. {newline}"
            f"The package is using PyTorch version: {torch.__version__} " + gpu)

if __name__ == "__main__":
    print(run())



# coding=utf-8
"""
File containing the abstract kernel classes.

@author: HENRI DE PLAEN
@copyright: KU LEUVEN
@license: MIT
@date: March 2021
"""
from __future__ import annotations
import torch
from torch import Tensor
from abc import ABCMeta, abstractmethod
from math import inf

from .. import utils
from ..feature.sample import Sample


@utils.extend_docstring(Sample)
class _BaseKernel(Sample, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        super(_BaseKernel, self).__init__(*args, **kwargs)

    @abstractmethod
    def __str__(self):
        pass

    # PROPERTIES
    @property
    @abstractmethod
    def explicit(self) -> bool:
        r"""
        True if the method has an explicit formulation, False otherwise.
        """
        pass

    @property
    def hparams_fixed(self):
        r"""
        Dictionnary containing the hyper-parameters and their values. This can be relevant for monitoring.
        """
        return {**super(_BaseKernel, self).hparams_fixed}

    @property
    def hparams_variable(self) -> dict:
        r"""
        Dictionnary containing the parameters and their values. This can be relevant for monitoring.
        """
        return {**super(_BaseKernel, self).hparams_variable}

    @property
    @abstractmethod
    def dim_feature(self) -> int | inf:
        r"""
        Returns the dimension of the explicit feature map if it exists.
        """
        pass

    # def merge_idxs(self, *args, **kwargs):
    #     raise NotImplementedError
    #     # self.dmatrix()
    #     # return torch.nonzero(torch.triu(self.dmatrix()) > (1 - kwargs["mtol"]), as_tuple=False)
    #
    # def merge(self, idxs):
    #     raise NotImplementedError
    #     # # suppress added up kernel
    #     # self._Sample = (self._Sample.gather(dim=0, index=idxs[:, 1]) +
    #     #                 self._Sample.gather(dim=0, index=idxs[:, 0])) / 2
    #
    #     self.dmatrix()
    #     # suppress added up kernel entries in the kernel matrix
    #     self._Cache["K"].gather(dim=0, index=idxs[:, 1], out=self._Cache["K"])
    #     self._Cache["K"].gather(dim=1, index=idxs[:, 1], out=self._Cache["K"])
    #
    # def reduce(self, idxs):
    #     raise NotImplementedError
    #     self._Sample.gather(dim=0, index=idxs, out=self._Sample)

    ###################################################################################################
    ################################### MATHS ARE HERE ################################################
    ###################################################################################################

    @abstractmethod
    def _implicit(self, x, y) -> Tensor:
        pass

    def _implicit_with_none(self, x=None, y=None) -> Tensor:
        # implicit raw
        if x is None:
            x = self.current_sample_projected
        if y is None:
            y = self.current_sample_projected
        return self._implicit(x, y)

    def _implicit_self(self, x=None):
        K = self._implicit_with_none(x, x)
        return torch.diag(K)

    @abstractmethod
    def _explicit(self, x) -> Tensor:
        pass

    def _explicit_with_none(self, x=None):
        # explicit raw
        if x is None:
            x = self.current_sample_projected
        return self._explicit(x)

    def _phi(self):
        r"""
        Returns the explicit feature map with default centering and normalization. If already computed, it is
        recovered from the cache.

        :param overwrite: By default, the feature map is recovered from cache if already computed. Force overwrites
            this if True., defaults to False.
        """
        if self.empty_sample:
            raise utils.NotInitializedError('The sample has not been initialized yet.')
        return self._get("phi", level_key="sample_phi", fun=self.explicit)

    def _C(self) -> Tensor:
        r"""
        Returns the covariance matrix with default centering and normalization. If already computed, it is recovered
        from the cache.

        :param overwrite: By default, the covariance matrix is recovered from cache if already computed. Force overwrites
            this if True., defaults to False
                """
        def fun():
            self._check_sample()
            scale = 1 / self.num_idx
            phi = self._phi()
            return scale * phi.T @ phi
        return self._get("C", level_key="sample_C", fun=fun)

    def _K(self, explicit=None, force : bool=False) -> Tensor:
        r"""
        Returns the kernel matrix with default centering and normalization. If already computed, it is recovered from
        the cache.

        :param explicit: Specifies whether the explicit or implicit formulation has to be used. Always uses the
            the explicit if available.
        :param force: By default, the kernel matrix is recovered from cache if already computed. Force overwrites
            this if True., defaults to False
        """
        def fun(explicit):
            self._check_sample()
            if explicit is None: explicit = self.explicit
            if explicit:
                phi = self._phi()
                return phi @ phi.T
            else:
                return self._explicit_with_none()
        return self._get("K", level_key="sample_K", fun=lambda: fun(explicit))

    # ACCESSIBLE METHODS
    def phi(self, x=None) -> Tensor:
        r"""
        Returns the explicit feature map :math:`\phi(\cdot)` of the specified points.

        :param x: The datapoints serving as input of the explicit feature map. If `None`, the sample will be used.,
            defaults to `None`
        :type x: Tensor(,dim_input), optional
        :raises: ExplicitError
        """
        if x is None:
            return self._phi()

        x = utils.castf(x)
        x = self.transform_input(x)
        return self._explicit_with_none(x)

    def k(self, x=None, y=None, explicit=None) -> Tensor:
        r"""
        Returns a kernel matrix, either of the sample, either out-of-sample, either fully out-of-sample.

        .. math::
            K = [k(x_i,y_j)]_{i,j=1}^{N,M},

        with :math:`\{x_i\}_{i=1}^N` the out-of-sample points (`x`) and :math:`\{y_i\}_{j=1}^N` the sample points
        (`y`).

        .. note::
            In the case of centered kernels, this computation is more expensive as it requires to _center according to
            the sample data, which implies computing a statistic on the out-of-sample kernel matrix and thus
            also computing it.

        :param x: Out-of-sample points (first dimension). If `None`, the default sample will be used., defaults to `None`
        :param y: Out-of-sample points (second dimension). If `None`, the default sample will be used., defaults to `None`

        :type x: Tensor(N,dim_input), optional
        :type y: Tensor(M,dim_input), optional

        :return: Kernel matrix
        :rtype: Tensor(N,M)

        :raises: ExplicitError
        """
        if x is None and y is None:
            return self._K(explicit=self.explicit)

        # in order to get the values in the correct format (e.g. coming from numpy)
        x = utils.castf(x)
        y = utils.castf(y)
        x = self.transform_input(x)
        y = self.transform_input(y)
        return self._implicit_with_none(x, y)

    def c(self, x=None) -> Tensor:
        r"""
        Out-of-sample explicit matrix.

        .. math::
            C = \frac1M\sum_{i}^{M} \phi(x_i)\phi(x_i)^\top.

        :param x: Out-of-sample points (first dimension). If `None`, the default sample will be used., defaults to `None`
        :type x: Tensor(N,dim_input), optional

        :return: Covariance matrix
        :rtype: Tensor(dim_feature,dim_feature)
        """

        phi = self.phi(x)
        scale = 1 / x.shape[0]
        return scale * phi.T @ phi

    def forward(self, x, representation="implicit") -> Tensor:
        """
        Passes datapoints through the kernel.

        :param x: Datapoints to be passed through the kernel.
        :param representation: Chosen representation. If `dual`, an out-of-sample kernel matrix is returned. If
            `primal` is specified, it returns the explicit feature map., defaults to `dual`

        :type x: Tensor(,dim_input)
        :type representation: str, optional

        :return: Out-of-sample kernel matrix or explicit feature map depending on `representation`.

        :raises: RepresentationError
        """

        def explicit(x):
            return self.phi(x)

        def implicit(x):
            return self.k(x)

        switcher = {"explicit": explicit,
                    "implicit": implicit}

        fun = switcher.get(representation, utils.RepresentationError(cls=self))
        return fun(x)

    @property
    def K(self) -> Tensor:
        r"""
        Returns the kernel matrix on the sample data. Same result as calling :py:func:`k()`, but faster.
        It is loaded from memory if already computed and unchanged since then, to avoid re-computation when recurrently
        called.

        .. math::
            K_{ij} = k(x_i,x_j).
        """
        return self._K()

    @property
    def C(self) -> Tensor:
        r"""
        Returns the explicit matrix on the sample datapoints.

        .. math::
            C = \frac{1}{\texttt{num_idx}}\sum_i^\texttt{num_idx} \phi(x_i)\phi(x_i)^\top.
        """
        return self._C()

    @property
    def Phi(self) -> Tensor:
        r"""
        Returns the explicit feature map :math:`\phi(\cdot)` of the sample datapoints. Same as calling
        :py:func:`phi()`, but slightly faster.
        It is loaded from memory if already computed and unchanged since then, to avoid re-computation when recurrently
        called.
        """
        return self._phi()

    ## DIRECT ATTRIBUTES

    @property
    def Cov(self) -> torch.Tensor:
        r"""
        Returns the covariance matrix of the sample. Same as calling self.cov().
        """
        return self.cov()

    @property
    def Corr(self) -> torch.Tensor:
        r"""
        Returns the correlation matrix of the sample. Same as calling self.corr().
        """
        return self.corr()

    def implicit_preimage(self, k_image: Tensor | None = None, method: str = 'knn', **kwargs):
        r"""
        Computes a pre-image of coefficients in the RKHS of the kernel, given by ``k_image``.
        Different methods are available:

        * ``'knn'``: Nearest neighbors. We refer to :py:func:`kerch.method.knn` for more details.
        * ``'smoother'``: Kernel smoothing. We refer to :py:func:`kerch.method.smoother` for more details
        * ``'iterative'``: Iterative optimization. We refer to :py:func:`kerch.method.iterative_preimage_k` for more details

        :param k_image: RKHS coefficients to be inverted. If not specified (``None``), the kernel matrix on the sample is used.
        :type k_image: torch.Tensor [num_points, num_idx], optional
        :param method: Pre-image method to be used. Defaults to ``'knn'``.
        :type method: str, optional
        :param \**kwargs: Additional parameters of the pre-image method used. Please refer to its documentation for
            further details.
        :type \**kwargs: dict, optional
        :return: Pre-image
        :rtype: torch.Tensor [num_points, dim_input]
        """

        # DEFENSIVE
        k_image = utils.castf(k_image)
        if k_image is None:
            k_image = self.K

        if torch.all(k_image < 0):
            self._logger.warning(f"The argument k_coefficient contains negative values, which should never be the case by "
                              f"definition of a RKHS.")

        # PRE-IMAGE
        method = method.lower()
        if method == 'knn':
            from ..method import knn
            return knn(dists=-k_image, observations=self.current_sample, **kwargs)
        elif method == 'smoother':
            from ..method import smoother
            return smoother(coefficients=k_image, observations=self.current_sample, **kwargs)
        elif method == 'iterative':
            from ..method import iterative_preimage_k
            return iterative_preimage_k(k_image=k_image, kernel=self, **kwargs)
        else:
            raise AttributeError('Unknown or non-implemented preimage method.')




# coding=utf-8
import torch
import logging

FTYPE = torch.float32
ITYPE = torch.int32

def gpu_available() -> bool:
    r"""
    Returns whether GPU-enhanced computation is possible and configured on this machine.
    """
    import torch.cuda
    if torch.cuda.is_available():
        from ..feature.logger import _GLOBAL_LOGGER
        _GLOBAL_LOGGER._logger.info("Using CUDA version " + torch.version.cuda)
        return True
    return False

def set_ftype(type):
    r"""
    Sets the generic floating type :attr:`kerch.FTYPE` used throughout the package. Typical choices include half precision
    :attr:`torch.float16`, single precision :attr:`torch.float32` (default) and double precision :attr:`torch.float64`.

    :param type: Default floating type to be used.
    :type type: PyTorch type

    .. warning:
        This does not affect the already instantiated tensors. It is thus preferable to set this in the beginning of
        the code to avoid any type mismatch.
    """

    assert isinstance(type, torch.dtype), 'The type is not an instance of torch.dtype.'
    global FTYPE
    FTYPE = type
    logging.warning('Changing name has to be carefully considered as changes '
                    'after initialization may lead to inconsistencies.')

def set_itype(type):
    r"""
    Sets the generic integer type :attr:`kerch.ITYPE` used throughout the package. Typical choices include short
    integers :attr:`torch.int16` (-32 768 to 32 767), classical integers :attr:`torch.int32` (-2^31-1 to
    2^31, default) and long integers :attr:`torch.int64` (-2^63-1 to 2^63). We do not advise on using unsigned integers
    because of their limited support in PyTorch.

    :param type: Default integer type to be used.
    :type type: PyTorch type

    .. warning:
        This does not affect the already instantiated tensors. It is thus preferable to set this in the beginning of
        the code to avoid any type mismatch.
    """
    assert isinstance(type, torch.dtype), 'The type is not an instance of torch.dtype.'
    global ITYPE
    ITYPE = type
    logging.warning('Changing name has to be carefully considered as changes '
                    'after initialization may lead to inconsistencies.')


def set_eps(eps: float):
    r"""
    Sets the generic epsilon value used throughout the toolbox to guarantee stability.

    :param eps: Default epsilon type to be used.
    :type eps: float

    .. warning:
        It is preferable to set this in the beginning of the code to avoid any type mismatch, preferably after setting
        the data type.
    """
    assert eps>=0, 'The EPS value has to be positive'
    global EPS
    EPS = torch.tensor(eps, dtype=FTYPE)
    return EPS

EPS = set_eps(1.e-7)


import torch
from torch import Tensor as T
import numpy as np

from . import utils
from ._logger import _Logger


class _DataHolder(_Logger):

    def __init__(self, **kwargs):
        super(_DataHolder, self).__init__(**kwargs)
        self._training_data = None
        self._training_labels = None
        self._validation_data = None
        self._validation_labels = None
        self._testing_data = None
        self._testing_labels = None

    def set_data(self, training_data=None, training_labels=None,
                 validation_data=None, validation_labels=None,
                 testing_data=None, testing_labels=None):
        r"""
        Sets the different datasets explicitly.
        """
        self._training_data = utils.castf(training_data)
        self._training_labels = utils.castf(training_labels)
        self._validation_data = utils.castf(validation_data)
        self._validation_labels = utils.castf(validation_labels)
        self._testing_data = utils.castf(testing_data)
        self._testing_labels = utils.castf(testing_labels)

        # not necessary:
        # if isinstance(self, _Sample):
        #     self.init_sample(self._training_data)

    def set_data_prop(self, data=None, labels=None, proportions=None) -> None:
        r"""
        Sets the different datasets based on proportions.
        """
        if proportions is None:
            proportions = [.7, .15, .15]
        else:
            assert len(proportions)==3, 'The proportions should contain 3 elements (training, validation and test). ' \
                                        'Please fill 0 if you do not want the create on of these sets'

        data_list, labels_list = _DataHolder._split_data(data, labels, proportions)

        self.set_data(training_data=data_list[0], training_labels=labels_list[0],
                      validation_data=data_list[1], validation_labels=labels_list[1],
                      testing_data=data_list[2], testing_labels=labels_list[2])

    def set_data_training_test(self, training_data=None, training_labels=None,
                                     testing_data=None, testing_labels=None,
                                     validation_prop=0.) -> None:
        r"""
        Sets the different datasets based on a training and a test set only. A validation set can be created from
        the training set based on some proportion.
        """
        data_list, labels_list = _DataHolder._split_data_two(training_data, training_labels, prop=validation_prop)
        training_data, validation_data = data_list
        training_labels, validation_labels = labels_list

        self.set_data(training_data, training_labels,
                      validation_data, validation_labels,
                      testing_data, testing_labels)

    def _get_fold(self, prop:float=.2):
        return _DataHolder._split_data_two(self._training_data, self._training_labels, prop=prop)

    def _get_default_data(self, data=None, labels=None):
        if data is None:
            data = self._training_data
            labels = self._training_labels
        elif labels is None: # if labels is None and not value
            raise NotImplementedError # reconstruction error for example in KPCA (still see how to implement it neatly).
        return data, labels

    ####################################################################################################################

    @staticmethod
    def _split_data(data, labels=None, props=None):
        r"""
        Splits the value in multiple random datasets based on props.
        """
        if props is None:
            return data, labels

        n = data.shape[0]
        perm = np.random.permutation(n)
        data = data[perm]
        if labels is not None:
            labels = labels[perm]

        data_list = []
        labels_list = []
        pos_start = 0
        for prop in props:
            if prop==0.:
                data_list.append(None)
                labels_list.append(None)
            else:
                pos_end = pos_start + round(prop * n)
                data_list.append(data[pos_start:pos_end])
                if labels is not None:
                    labels_list.append(labels[pos_start:pos_end])
                else:
                    labels_list.append(None)

        return data_list, labels_list

    @staticmethod
    def _split_data_two(data, labels=None, prop:float=.2):
        return _DataHolder._split_data(data, labels, props=[1. - prop, prop])



# coding=utf-8
def reverse_dict(val: dict) -> dict:
    r"""
    Inverts to keywords and the values in a dictionnary.

    :param val: Dictionary to be inverted.
    :return: Inverted dictionnary.

    :type val: dict
    :rtype: dict
    """
    return {v: k for k, v in val.items()}


import torch

def knn(coefficients: torch.Tensor, x: torch.Tensor, num: int = 1) -> torch.Tensor:
    r"""
    Returns the sum of the num closest x, the distance being given by the coefficients.
    """
    num_points, num_coefficients = coefficients.shape[0]
    num_x = x.shape[0]

    assert num_coefficients == num_x, \
        f'KNN: Incorrect number of coefficients ({num_coefficients}), ' \
        f'compared to the number of points ({num_x}).'

    preimages = []
    for idx in range(num_points):
        _, indices = torch.sort(coefficients, descending=True)
        loc_sol = torch.mean(x[indices[:num]], dim=0)
        preimages.append(loc_sol)
    return torch.vstack(preimages)


import torch


def invert_dict(name, dict):
    # The name becomes the key inside the key and the original key become the names.
    # The output is of the form [(new_name1, new_dict1), (new_name2, new_dict2), ...].
    dicts = []
    for key in dict:
        new_dict = {name: dict[key]}
        dicts.append((key, new_dict))
    return dicts


def add_dict(name, dict):
    dicts = {}
    for key in dict:
        new_entry = {name + ' ' + key: dict[key].sample}
        dicts = {**new_entry, **dicts}
    return dicts


def verify_dim(M):
    return M


def process_y(y):
    if len(list(y.shape)) == 1:
        y = y.unsqueeze(1)
    return y



# coding=utf-8
import torch
from torch import Tensor as T
from abc import ABCMeta, abstractmethod

from ._View import _View
from .. import utils


class _Level(_View, metaclass=ABCMeta):
    r"""
    A level consists in a view (kernel + weight/hidden) with various optimization attributes.

    :param eta: level weight :math:`\eta`., defaults to 1.
    :param representation: Chosen representation, "primal" or "dual"., defaults to "dual".

    :type eta: double, optional
    :type representation: str, optional
    """

    @utils.extend_docstring(_View)
    def __init__(self, *args, **kwargs):
        super(_Level, self).__init__(*args, **kwargs)
        self._eta = kwargs.pop('eta', 1.)
        self.watch_properties = kwargs.pop('watch_properties', [])
        self._parameter_related_cache = []

        # INITIALIZE THE PARAMETERS WHEN POSSIBLE
        try:
            self.init_parameters()
        except AssertionError:
            pass

    @property
    def hparams_fixed(self) -> dict:
        return {'Level type': self.__class__.__name__,
                'Level eta': self.eta,
                **super(_Level, self).hparams_fixed}

    @property
    def hparams_variable(self) -> dict:
        return {**super(_Level, self).hparams_variable}

    @property
    def watch_properties(self) -> list[str]:
        r"""
        Properties to be watched (monitored). Their values can be accessed through the property `watched_properties`.
        This is relevant e.g. in case of training.
        """
        return self._watch_properties

    @watch_properties.setter
    def watch_properties(self, val: list[str]):
        assert isinstance(val, list), 'The watch argument must be a list of strings.'
        for name in val:
            assert name in dir(self), f"The watch element {name} does not exist for this level."
        self._watch_properties = val

    @property
    def eta(self) -> float:
        r"""
        Level weight :math:`\eta`, e.g. for the weight of the loss.
        """
        return self._eta

    @property
    def _I_primal(self) -> T:
        r"""
        To avoid multiple reinitializations, certainly when working on GPU, the value is stored in memory the first
        time it is called, to be re-used later.
        """
        level_key = "Level_I_default_representation" if self._representation == 'primal' \
            else "Level_I_other_representation"
        fun = lambda: torch.eye(self.dim_feature, dtype=utils.FTYPE, device=self._param_device)
        _I_primal = self._get("Level_I_primal", level_key=level_key, fun=fun, persisting=True)
        if _I_primal.shape[0] != self.dim_feature:
            _I_primal = self._get("Level_I_primal", level_key=level_key, fun=fun, overwrite=True, persisting=True)
        return _I_primal

    @property
    def _I_dual(self) -> T:
        r"""
        To avoid multiple reinitializations, certainly when working on GPU, the value is stored in memory the first
        time it is called, to be re-used later.
        """
        level_key = "Level_I_default_representation" if self._representation == 'dual' \
            else "Level_I_other_representation"
        fun = lambda: torch.eye(self.num_idx, dtype=utils.FTYPE, device=self._param_device)
        _I_dual = self._get("Level_I_dual", level_key=level_key, fun=fun, persisting=True)
        if _I_dual.shape[0] != self.num_idx:
            _I_dual = self._get("Level_I_dual", level_key=level_key, fun=fun, overwrite=True, persisting=True)
        return _I_dual

    @abstractmethod
    def _solve_primal(self) -> None:
        r"""
        Solves the dual formulation on the sample.
        """
        pass

    @abstractmethod
    def _solve_dual(self) -> None:
        r"""
        Solves the primal formulation on the sample.
        """
        pass

    @torch.no_grad()
    def solve(self, sample=None, target=None, representation=None, **kwargs) -> None:
        r"""
        Fits the model according to the input ``sample`` and output ``target``. Many models have both a primal and
        a dual formulation to be fitted.

        :param representation: Representation of the model (``"primal"`` or ``"dual"``)., defaults to ``"dual"``.
        :type representation: str, optional
        """

        # self._log.debug("The fitting is always done on the full sample data, regardless of the stochastic state.")
        # set the sample to input (always works for the underlying kernel)

        if sample is not None:
            self.init_sample(sample)
        if target is not None:
            self.target = target

        # verify that the sample has been initialized
        if self._num_total is None:
            self._logger.error("Cannot solve as no input has been provided nor a sample already exists")
            return

        # check the representation is correct and set it to the default Level value if None
        representation = utils.check_representation(representation, default=self._representation, cls=self)

        # execute the corresponding fitting
        switcher = {"primal": self._solve_primal,
                    "dual": self._solve_dual}

        switcher.get(representation)()

    ####################################################################################################################

    @abstractmethod
    def loss(self, representation=None) -> T:
        pass

    def losses(self, representation=None) -> dict:
        r"""
        Different components of the losses.
        """
        return {'Level loss': self.loss().data.detach().cpu().item()}

    @property
    def watched_properties(self) -> dict:
        r"""
        A dictionary containing the values of the properties that are specified in `watch_properties`.
        """
        props = dict()
        for name in self.watch_properties:
            val = getattr(self, name)
            if isinstance(val, T):
                val = val.data.detach().cpu()
            props[name] = val
        return props

    def before_step(self) -> None:
        r"""
        Performs steps required before each training step.
        """
        losses_keys = self.losses().keys()
        self._remove_from_cache(losses_keys)

    def after_step(self) -> None:
        r"""
        Performs after-step operations, for example a transform of the parameters onto some manifold.
        """
        self._remove_from_cache(self._parameter_related_cache)



"""
Abstract RKM View class.

@author: HENRI DE PLAEN
@copyright: KU LEUVEN
@license: MIT
@date: March 2021, rewritten in June 2022
"""

import torch
from torch import Tensor
from math import sqrt

import kerch.rkm
from kerch import utils
from kerch.kernel import factory, _Base
from kerch._sample import _Sample
from ._view import _View


@utils.extend_docstring(_Sample)
class View(_View, _Sample):
    r"""
    :param kernel: Initiates a View _Based on an existing kernel object. If the value is not `None`, all other
        parameters are neglected and inherited from the provided kernel., default to `None`
    :param bias: Bias
    :param bias_trainable: defaults to `False`

    :type kernel: kerch.kernel._Projected, optional
    :type bias: bool, optional
    :type bias_trainable: bool, optional
    """

    @utils.kwargs_decorator({
        "kernel": None,
        "bias": False,
        "bias_trainable": False,
        "kappa": 1.,
        "targets" : None
    })
    def __init__(self, *args, **kwargs):
        """
        A View is made of a kernel and primal or dual variables. This second part is handled by the daughter classes.
        """
        super(View, self).__init__(*args, **kwargs)

        # KAPPA
        self._kappa = kwargs["kappa"]
        self._kappa_sqrt = sqrt(self._kappa)
        self._mask = None

        # BIAS
        self._bias_trainable = kwargs["bias_trainable"]
        self._requires_bias = kwargs["bias"]
        self._bias = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE),
                                        requires_grad=self._bias_trainable)


        # KERNEL INIT
        # always share sample with kernel
        kernel = kwargs["kernel"]
        if kernel is None:
            self._kernel = factory(**{**kwargs,
                                      "sample": self.sample,
                                      "idx_sample": self.idx})
        elif isinstance(kernel, _Base):
            self._log.info("Initiating View _Based on existing kernel and overwriting its sample.")
            self._kernel = kernel
            self._kernel.init_sample(sample=self.sample,
                                     idx_sample=self.idx)
        else:
            raise TypeError("Argument kernel is not of the kernel class.")

        # TARGETS
        targets = kwargs["targets"]
        targets = utils.castf(targets, tensor=True)
        if self._dim_output is None and targets is not None:
            self._dim_output = targets.shape[1]
        self.targets = targets



        self._log.debug("View initialized with " + str(self._kernel))

    def __str__(self):
        return "view with " + str(self._kernel)

    def _reset_weight(self) -> None:
        self._weight = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE,
                                                      device=self._weight.device),
                                          requires_grad=self._weight.requires_grad)

    @property
    def kappa(self) -> float:
        return self._kappa

    @kappa.setter
    def kappa(self, val:float):
        self._kappa = val

    @property
    def bias(self) -> Tensor:
        if self._bias.nelement() != 0:
            return self._bias
        self._log.debug("No bias has been initialized yet.")

    @bias.setter
    def bias(self, val):
        if val is not None:
            val = utils.castf(val, dev=self._bias.device).squeeze()
            dim_val = len(val.shape)

            # verifying the shape of the bias
            if dim_val == 0:
                val = val.repeat(self._dim_output)
            elif dim_val > 1:
                self._log.error("The bias can only be set to a scalar or a vector. "
                                "This operation is thus discarded.")
            # setting the value
            if self._bias.nelement() == 0:
                self._bias = torch.nn.Parameter(val,
                                                requires_grad=self._bias_trainable)
            else:
                self._bias.data = val
                # zeroing the gradients if relevant
                if self._bias_trainable:
                    self._bias.grad.sample.zero_()

    @property
    def bias_trainable(self) -> bool:
        return self._bias_trainable

    @bias_trainable.setter
    def bias_trainable(self, val: bool):
        self._bias_trainable = val
        self._bias.requires_grad = self._bias_trainable

    @property
    def _bias_exists(self) -> bool:
        return self._bias.nelement() != 0

    ####################################################################################################################
    ## TARGETS

    @property
    def targets(self) -> Tensor:
        r"""
            Targets to be matched to.
        """
        if self._targets is None:
            self._log.warning("Empty target values.")
        return self._targets

    @targets.setter
    def targets(self, val):
        val = utils.castf(val, dev=self._sample.device, tensor=True)
        if val is None:
            self._log.debug("Targets set to empty values.")
        else:
            assert self.dim_output == val.shape[1], f"The shape of the given target {val.shape[1]} does not match the" \
                                                    f" required one {self.dim_output}."
            assert self.num_sample == val.shape[0], f"The number of target points {val.shape[0]} does not match the " \
                                                    f"required one {self.dim_input}."
        self._targets = torch.nn.Parameter(val)

    @property
    def current_targets(self) -> Tensor:
        r"""
            Returns the targets that are currently used in the computations and for the normalizing and centering
            statistics if relevant.
        """
        return self.targets[self.idx, :]


    ########################################################################
    @property
    def dim_feature(self) -> int:
        return self.kernel.dim_feature

    @property
    def kernel(self) -> _Base:
        r"""
        The kernel used by the model or View.
        """
        return self._kernel

    def set_kernel(self, val: _Base):
        r"""
        For some obscure reason, this does not work as a setter (@kernel.setter).
        TODO: find out why and solve
        """
        self._log.info("Updating View _Based on an external kernel and overwriting its sample.")
        self._kernel = val
        self._kernel.init_sample(sample=self.sample,
                                 idx_sample=self.idx)

    #########################################################################
    ## WEIGHT
    def _update_weight_from_hidden(self):
        if self._hidden_exists:
            # will return a ExplicitError if not available
            self.weight = self.Phi.T @ self.H
            self._log.debug("Setting the weight _Based on the hidden values.")
        else:
            self._log.info("The weight cannot _Based on the hidden values as these are unset.")

    def _update_hidden_from_weight(self):
        raise NotImplementedError

    ## MATHS

    def phi(self, x=None) -> Tensor:
        return self._kappa_sqrt * self.kernel.phi(x)

    def k(self, x=None) -> Tensor:
        return self.kappa * self.kernel.k(x)

    @property
    def K(self) -> Tensor:
        return self.kappa * self.kernel.K

    def forward(self, x=None, representation=None):
        representation = utils.check_representation(representation, default=self._representation)
        if self._bias_exists:
            return self.phiw(x, representation) + self._kappa_sqrt * self._bias[:, None]
        else:
            return self.phiw(x, representation)



# coding=utf-8
from __future__ import annotations
from abc import ABCMeta

from typing import Iterator
import torch

from ..level import Level, factory
from ..utils import NotInitializedError, kwargs_decorator
from ..feature.stochastic import Stochastic


class Model(Stochastic, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)
        self._levels: list[Level] = list()

    @property
    def levels(self) -> Iterator[Level]:
        yield from self._levels

    @property
    def num_levels(self) -> int:
        return len(self._levels)

    def _check_levels(self):
        if self.num_levels == 0:
            raise NotInitializedError(cls=self, message="The model does not contain any level.")

    def level(self, num: int) -> Level:
        assert num < self.num_levels, f"Model has less levels ({self.num_levels}) than requested ({int})."
        return next(x for i, x in enumerate(self.levels) if i == num)

    @property
    def _first_level(self) -> Level:
        self._check_levels()
        return next(self.levels)

    @property
    def _last_level(self) -> Level:
        self._check_levels()
        return self._levels[-1]

    @property
    def dim_input(self) -> int:
        return self._first_level.dim_input

    @property
    def dim_output(self) -> int:
        return self._last_level.dim_output

    @property
    def num_sample(self) -> int:
        return self._first_level.num_sample

    @property
    def hparams_fixed(self) -> dict:
        val = dict()
        for num, level in enumerate(self.levels):
            name = f"[LEVEL{num} {level.__class__.__name__}] "
            val = {**val, **{name + key: val for key, val in level.hparams_fixed.items()}}
        return val

    @property
    def hparams_variable(self) -> dict:
        val = dict()
        for num, level in enumerate(self.levels):
            name = f"[LEVEL{num} {level.__class__.__name__}] "
            val = {**val, **{name + key: val for key, val in level.hparams_variable.items()}}
        return val

    @property
    def losses(self) -> dict:
        val = dict()
        for num, level in enumerate(self.levels):
            name = f"[LEVEL{num} {level.__class__.__name__}] "
            val = {**val, **{name + key: val for key, val in level.losses().items()}}
        return val

    @property
    def watched_properties(self) -> dict:
        val = dict()
        for num, level in enumerate(self.levels):
            name = f"[LEVEL{num} {level.__class__.__name__}] "
            val = {**val, **{name + key: val for key, val in level.watched_properties.items()}}
        return val

    def init_sample(self, sample=None):
        self._first_level.init_sample(sample=sample)
        self._num_total = self._first_level.num_sample
        self.stochastic(idx=self._first_level.idx)

    @property
    def empty_sample(self) -> bool:
        return self._first_level.empty_sample

    def init_levels(self, full: bool = False) -> None:
        r"""
        Initializes all levels based on the model sample.

        :param full: If specified to False, the sample initialization is going to randomly initialized for the deeper
            levels to avoid running a forward on the full data. This is relevant in a stochastic setting. Defaults
            to True.
        :type full: bool, optional
        """
        self._first_level.init_parameters(overwrite=False)
        if full:
            x = self._first_level()
            for level in self._levels[1:]:
                level.init_sample(sample=x)
                level.init_parameters(overwrite=False)
                x = level()
        else:
            for level in self._levels[1:]:
                level.num_sample = self.num_sample
                level.init_parameters(overwrite=False)

    def init_target(self, target: torch.Tensor, num_level: int = -1) -> None:
        if num_level == -1:
            level = self._last_level
        else:
            level = self.level(num_level)
        try:
            level.target = target
        except NotInitializedError:
            raise NotInitializedError(cls=self, message="Please initialize the sample and the levels first.")

    def forward(self, x: torch.Tensor | None = None) -> torch.Tensor:
        if self.training:
            for level in self.levels:
                if x is None:
                    level.stochastic(idx=self._idx_stochastic)
                else:
                    level.update_sample(x, idx_sample=self._idx_stochastic)
                if level.level_trainable is False:
                    level.solve()
                x = level()
        else:
            for level in self.levels:
                x = level(x)
        return x

    def loss(self):
        return sum([level.eta * level.loss() for level in self.levels])

    def stochastic(self, idx=None, prop=None):
        super().stochastic(idx, prop)
        for level in self.levels:
            level.stochastic(idx=self._idx_stochastic)

    def before_step(self):
        for level in self.levels:
            level.before_step()

    def after_step(self):
        for level in self.levels:
            level.after_step()

    def _stiefel_parameters(self, recurse=True) -> Iterator[torch.nn.Parameter]:
        for level in self.levels:
            yield from level._stiefel_parameters(recurse=recurse)

    def _euclidean_parameters(self, recurse=True) -> Iterator[torch.nn.Parameter]:
        for level in self.levels:
            yield from level._euclidean_parameters(recurse=recurse)

    def _slow_parameters(self, recurse=True) -> Iterator[torch.nn.Parameter]:
        for level in self.levels:
            yield from level._slow_parameters(recurse=recurse)

    def train(self, mode=True):
        super().train(mode)
        for level in self.levels:
            level.train(mode)

    @kwargs_decorator({"level_type": None,
                       "constraint": "soft"})
    def _append_level(self, *args, **kwargs) -> None:
        # create level and remove sample
        level = factory(*args, **kwargs)

        # verify the correct assignment of the output dimension
        try:
            _ = level.dim_output
        except NotInitializedError:
            raise AssertionError("The argument dim_output is not specified. This is not required if targets are "
                                 "explicitly specified during initialization.")

        # verify the correct assignment of the input dimension
        try:
            _ = level.dim_input
        except NotInitializedError:
            if self.num_levels > 0:
                level.dim_input = self.level(self.num_levels - 1).dim_output




        # overwrite param_trainable
        match kwargs['constraint']:
            case "soft":
                level.level_trainable = True
            case "hard":
                level.level_trainable = False
            case _:
                raise NameError('The constraint argument must be either soft or hard.')



        # verify dimension consistency
        if self.num_levels > 0:
            assert self.level(self.num_levels - 1).dim_output == level.dim_input, \
                f"The input dimension of the level to be added ({level.dim_input}) does not correspond to the " \
                f"output dimension of the previous level ({self.level(self.num_levels - 1).dim_output})."

        self._levels.append(level)

    def reset(self, recurse=False, reset_persisting=True) -> None:
        for level in self.levels:
            level.reset(recurse=recurse, reset_persisting=reset_persisting)



# coding=utf-8
from .single_view import (KPCA as KPCA,
                          LSSVM as LSSVM,
                          PPCA as PPCA,
                          Ridge as Ridge,
                          Level as Level)
from .multi_view import (MVKPCA as MVKPCA)
from .factory import factory as factory


# coding=utf-8
__version__ = "0.3.1.4"
__author__ = "HENRI DE PLAEN"
__credits__ = "KU Leuven"
__status__ = "beta"
__date__ = "May 2024"
__license__ = "LGPL-3.0"

# GLOBAL MODULE-WIDE VARIABLES
_GLOBALS = {"PLOT_ENV": None,
            "LOG_LEVEL": 30,  # this corresponds to logging.WARNING
            }

__all__ = ['__version__', '__author__', '__credits__', '__status__', '__date__', '__license__',
           'kernel', 'level', 'model', 'data', 'train', 'opt', 'set_logging_level', 'get_logging_level',
           'gpu_available',
           'set_ftype', 'set_itype', 'DEFAULT_KERNEL_TYPE', 'DEFAULT_CACHE_LEVEL', 'FTYPE', 'ITYPE']

# IMPORTS
from . import kernel as kernel  # ok (tested & documented)
from . import level as level  # beta
from . import model as model  # beta
from . import data as data  # beta
from . import train as train  # alpha
from . import monitor as monitor  # beta
from . import plot as plot  # alpha
from . import opt as opt  # beta
from . import script as script  # alpha
from . import method as method  # alpha
from .feature.logger import (set_logging_level as set_logging_level,
                             get_logging_level as get_logging_level)
from .utils import (gpu_available as gpu_available,
                    FTYPE as FTYPE,
                    ITYPE as ITYPE,
                    set_ftype as set_ftype,
                    set_itype as set_itype,
                    set_eps as set_eps,
                    DEFAULT_KERNEL_TYPE as DEFAULT_KERNEL_TYPE,
                    DEFAULT_CACHE_LEVEL as DEFAULT_CACHE_LEVEL)



# coding=utf-8
"""
File containing the RBF kernel class.

@author: HENRI DE PLAEN
@copyright: KU LEUVEN
@license: MIT
@date: March 2021
"""
from __future__ import annotations
from abc import ABCMeta, abstractmethod
from .. import utils
from .kernel import Kernel
from torch import Tensor as T
from math import inf

from abc import ABCMeta
from ..utils import ExplicitError, extend_docstring


@extend_docstring(Kernel)
class Implicit(Kernel, metaclass=ABCMeta):

    def __init__(self, *args, **kwargs):
        super(Implicit, self).__init__(*args, **kwargs)

    def __str__(self):
        return f"Implicit kernel."

    @property
    def explicit(self) -> bool:
        return False

    @property
    def dim_feature(self) -> int | inf:
        r"""
        For implicit kernels, the feature dimension is infinite.
        """
        return inf

    def _explicit(self, x):
        raise ExplicitError(cls=self)

    def explicit_preimage(self, phi_image: T | None = None, method: str = 'explicit', **kwargs):
        r"""

        .. note::
            Not available for kernels that have no explicit feature map representation.


        """
        raise ExplicitError(cms=self)

    @abstractmethod
    def _implicit(self, x, y):
        pass




# coding=utf-8
from __future__ import annotations

from .Watcher import Watcher
from ..utils import extend_docstring


@extend_docstring(Watcher)
class Saver(Watcher):
    def __init__(self, *args):
        super(Saver, self).__init__(*args)

    def update(self,
               epoch: int,
               objective_loss: float,
               training_error: float | None = None,
               validation_error: float | None = None,
               test_error: float | None = None
               ) -> None:
        if epoch % self._num_epochs_save == 0:
            self.save_model(epoch)



import torch


def smoother(coefficients: torch.Tensor, x: torch.Tensor, num: int = 1) -> torch.Tensor:
    r"""
    Returns a weighted sum of x by the coefficients.
    """
    num_points, num_coefficients = coefficients.shape
    num_x = x.shape[0]

    assert num_coefficients == num_x, \
        f'KNN: Incorrect number of coefficients ({num_coefficients}), ' \
        f'compared to the number of points ({num_x}).'

    if num >= num_coefficients:
        return torch.einsum('ni,ij->nj', coefficients / torch.sum(coefficients, dim=0), x)

    preimages = []
    for idx in range(num_points):
        sorted_coefficients, indices = torch.sort(coefficients[idx, :], descending=True)
        nearest_coefficients = sorted_coefficients[:num]

        normalized_coefficients = nearest_coefficients / torch.sum(nearest_coefficients)
        loc_sol = torch.einsum('i,ij->j', normalized_coefficients, x[indices[:num], :])
        preimages.append(loc_sol)
    return torch.vstack(preimages)



"""
This abstract class has only one purpose, add a self._log property meant to log various information across the toolbox.
Doing it this way allows to get the name of the class instantiated and print more precise messages.

Author: HENRI DE PLAEN
Date: June 2022
"""

import logging
from abc import ABCMeta, abstractmethod

from . import _GLOBALS, utils

_LOGGING_FORMAT = "Kerch %(levelname)s [%(name)s]: %(message)s [%(pathname)s:%(lineno)d]"

_kerch_format = logging.Formatter(_LOGGING_FORMAT)
_kerch_handler = logging.StreamHandler()
_kerch_handler.setFormatter(_kerch_format)


class _Logger(metaclass=ABCMeta):
    r"""
    :param log_level: The logging level of this specific instance. If None, it will be the default logging level set
        for the toolbox., defaults to None.
    :type log_level: str or int
    """
    @utils.kwargs_decorator({
        "log_level": None,
        "name": None
    })
    def __init__(self, *args, **kwargs):
        self._name = kwargs["name"]
        class_name = self.__class__.__name__
        if self._name is not None and class_name != "_Logger":
            log_name = self._name + ' ' + class_name
        elif self._name is not None:
            log_name = self._name
        else:
            log_name = class_name
        self._log = logging.getLogger(name=log_name)
        self._log.addHandler(_kerch_handler)
        self.set_log_level(kwargs["log_level"])

    def set_log_level(self, level: int = None) -> int:
        r"""
        Sets a specific log Level to this object. It serves as a way to use specific log Level for a specific class,
        different than the current general kerch log Level.

        :param level: If the value is ``None``, the current general kerch log Level will be used (WARNING if not
            specified otherwise)., defaults to ``None``.
        :param name: int, optional
        """
        if level is None:
            level = _GLOBALS["LOG_LEVEL"]
        self._log.setLevel(level)
        return level

    def get_log_level(self) -> int:
        r"""
        Returns the log Level used by this object.
        """
        return self._log.level

    @property
    def name(self):
        r"""
        Name of the module. This is relevant in some applications
        """
        if self._name is not None:
            return self._name
        raise AttributeError

    @name.setter
    def name(self, val: str):
        self._log.error("The name cannot be changed after initialization.")


_GLOBAL_LOGGER = _Logger(name="global")


def set_log_level(level: int):
    r"""
    Sets the logging Level of the toolbox. The best is to use a value of the logging module.
    For example

    .. code-block::
        import kerch
        import logging

        kerch.set_log_level(logging.DEBUG)


    .. note::
        Changing the log value does not affect the already instantiated objects.


    """
    _GLOBALS["LOG_LEVEL"] = level
    _GLOBAL_LOGGER.set_log_level(level)


def get_log_level() -> str:
    r"""
    Returns the logging Level of the toolbox.
    """
    return logging.getLevelName(_GLOBALS["LOG_LEVEL"])



# coding=utf-8
from .datasets import (TwoMoons as TwoMoons,
                       PimaIndians as PimaIndians,
                       Spiral as Spiral,
                       Gaussians as Gaussians,
                       Iris as Iris)
from .factory import factory as factory


# coding=utf-8
"""
Allows for a sample set manager, with stochastic support.
"""

from typing import Iterator
import torch

from abc import ABCMeta, abstractmethod
from torch import Tensor
from typing import Union

from kerch import utils
from .cache import Cache
from .logger import Logger
from .stochastic import Stochastic
from ..transform import TransformTree
from ..utils import NotInitializedError


@utils.extend_docstring(Stochastic)
class Sample(Stochastic,  # manager stochastic indices
             metaclass=ABCMeta):
    r"""
    :param sample: Sample points used to compute the kernel matrix. When an out-of-sample computation is asked, it will
        be given relative to these samples., defaults to `None`
    :param sample_trainable: `True` if the gradients of the sample points are to be computed. If so, a graph is
        computed and the sample can be updated. `False` just leads to a static computation., defaults to `False`
    :param num_sample: Number of sample points. This parameter is neglected if `sample` is not `None` and overwritten by
        the number of points contained in sample., defaults to 1
    :param dim_input: Dimension of each sample point. This parameter is neglected if `sample` is not `None` and
        overwritten by the dimension of the sample points., defaults to 1

    :type sample: Tensor(num_sample, dim_input), optional
    :type sample_trainable: bool, optional
    :type num_sample: int, optional
    :type dim_input: int, optional

    :param idx_sample: Initializes the indices of the samples to be updated. All indices are considered if both
        `idx_stochastic` and `prop_stochastic` are `None`., defaults to `None`
    :param prop_sample: Instead of giving indices, specifying a proportion of the original sample set is also
        possible. The indices will be uniformly randomly chosen without replacement. The value must be chosen
        such that :math:`0 <` `prop_stochastic` :math:`\leq 1`. All indices are considered if both `idx_stochastic` and
        `prop_stochastic` are `None`., defaults to `None`.

    :type idx_sample: int[], optional
    :type idx_sample: float, optional

    :param sample_transform: TODO
    :type sample_transform: List[str]
    """

    _cache_elements = ["sample_transform"]

    def __init__(self, *args, **kwargs):
        super(Sample, self).__init__(*args, **kwargs)

        sample = kwargs.pop('sample', None)
        if sample is not None:
            sample = utils.castf(sample)
            self._num_total, self._dim_input = sample.shape
        else:
            self._dim_input = kwargs.pop('dim_input', None)
            self._num_total = kwargs.pop('num_sample', None)

        self._sample = torch.nn.Parameter(torch.empty((0, 0)))
        self._sample_trainable = kwargs.pop('sample_trainable', False)
        self._default_sample_transform = kwargs.pop('sample_transform', [])

        self.init_sample(sample, idx_sample=kwargs.pop('idx_sample', None), prop_sample=kwargs.pop('prop_sample', None))


    @property
    def dim_input(self) -> int:
        r"""
        Dimension of each datapoint.
        """
        if self._dim_input is None:
            raise NotInitializedError(cls=self, message="The input dimension has been initialized yet. "
                                                        "Please initialize it explicitly or assign a sample to "
                                                        "determine it.")
        return self._dim_input

    @dim_input.setter
    def dim_input(self, val: int):
        assert self._dim_input is None, "Cannot set the dimension of the sample points after initialization if the " \
                                        "sample data. Use init_sample() instead."
        self._dim_input = val
        if self._num_total is not None:
            self.init_sample()

    @property
    def empty_sample(self) -> bool:
        r"""
        Boolean specifying if the sample is empty or not.
        """
        return self._sample.nelement() == 0

    def _check_sample(self):
        if self.empty_sample:
            raise utils.NotInitializedError(cls=self, message='The sample has not been initialized yet.')

    @property
    def num_sample(self) -> int:
        r"""
        Number of datapoints in the sample set.
        """
        self._check_sample()
        return self._num_total

    @num_sample.setter
    def num_sample(self, val: int):
        assert self._num_total is None, "Cannot set the number of sample points after initialization of the " \
                                        "sample data. Use init_sample() instead."
        self._num_total = val
        if self._dim_input is not None:
            self.init_sample()

    @property
    def sample(self) -> torch.nn.Parameter:
        r"""
        Full original raw sample without any transform or potential stochastic selection.
        """
        self._check_sample()
        return self._sample

    @sample.setter
    def sample(self, val):
        assert val is not None, "Cannot assign the sample to None. Please use init_sample() if you want to " \
                                "re-initialize it."
        self.init_sample(val)

    @property
    def sample_trainable(self) -> bool:
        r"""
        Boolean if the sample data can be trained.
        """
        return self._sample_trainable

    @sample_trainable.setter
    def sample_trainable(self, val: bool):
        self._sample_trainable = val
        self._sample.requires_grad = self._sample_trainable

    @property
    def current_sample_projected(self) -> Tensor:
        r"""
        Returns the sample that is currently used in the computations and for the normalizing and centering statistics
        if relevant.
        """
        return self.sample_transform.projected_sample

    @property
    def current_sample(self) -> Tensor:
        return self._sample[self.idx, :]

    @torch.no_grad()
    def init_sample(self, sample=None, idx_sample=None, prop_sample=None):
        r"""
        Initializes the sample set (and the stochastic indices).

        :param sample: Sample points used for the various computations. When an out-of-sample computation is asked, it
            will be given relative to these samples. In case of overwriting a current sample, `num_sample` and
            `dim_input` are also overwritten. If `None` is specified, the sample data will be initialized according
            to `num_sample` and `dim_input` specified during the construction. If a previous sample set has been used,
            it will keep the same dimension by consequence. A last case occurs when `sample` is of the class
            `torch.nn.Parameter`: the sample will then use those values, and they can thus be shared with the level
            calling this method., defaults to `None`
        :type sample: Tensor, optional
        :param idx_sample: Initializes the indices of the samples to be updated. All indices are considered if both
            `idx_sample` and `prop_sample` are `None`., defaults to `None`
        :type idx_sample: int[], optional
        :param prop_sample: Instead of giving indices, specifying a proportion of the original sample set is also
            possible. The indices will be uniformly randomly chosen without replacement. The value must be chosen
            such that :math:`0 <` `prop_sample` :math:`\leq 1`. All indices are considered if both `idx_sample` and
            `prop_sample` are `None`., defaults to `None`.
        """
        sample = utils.castf(sample)

        if sample is None:
            self._logger.debug("Initializing new sample with the sample dimensions.")
            if self._num_total is None or self.dim_input is None:
                self._logger.info(
                    'The sample cannot be initialized because no sample data has been provided nor the '
                    'sample dimensions have been initialized yet.')
                return
            self._sample = torch.nn.Parameter(
                torch.nn.init.orthogonal_(torch.empty((self._num_total, self._dim_input),
                                                      dtype=utils.FTYPE,
                                                      device=self._sample.device)),
                requires_grad=self._sample_trainable)
        elif isinstance(sample, torch.nn.Parameter):
            self._logger.debug("Using existing sample defined as an external parameter.")
            self._num_total, self._dim_input = sample.shape
            self._sample_trainable = sample.requires_grad
            self._sample = sample
        else:
            self._logger.debug("Assigning new sample points based on the given values.")
            self._num_total, self._dim_input = sample.shape
            self._sample = torch.nn.Parameter(sample.data,
                                              requires_grad=self._sample_trainable)

        self.stochastic(idx=idx_sample, prop=prop_sample) # also resets

    @torch.no_grad()
    def update_sample(self, sample_values, idx_sample=None):
        r"""
        Updates the sample set. In contradiction to `init_samples`, this only updates the values of the sample and sets
        the gradients of the updated values to zero if relevant.

        :param sample_values: Values given to the updated samples.
        :param idx_sample: Indices of the samples to be updated. All indices are considered if `None`., defaults to
            `None`
        :type sample_values: Tensor
        :type idx_sample: int[], optional
        """

        if self.empty_sample:
            self._logger.warning('Cannot update the sample values of a None sample data.')
            return

        sample_values = utils.castf(sample_values, dev=self._sample.device)

        # use all indices if unspecified
        if idx_sample is None:
            idx_sample = self._all_idx

        # check consistency of indices
        assert len(idx_sample) == sample_values.shape[0], f"Number of sample values ({sample_values.shape[0]}) and " \
                                                          f"corresponding indices ({len(idx_sample)}) are not " \
                                                          f"consistent."
        # update the values
        self._sample.data[idx_sample, :] = sample_values.data

        # zeroing relevant gradient if relevant
        if self._sample_trainable:
            self._sample.grad.data[idx_sample, :].zero_()

        self._reset_cache(reset_persisting=False)

    def _euclidean_parameters(self, recurse=True) -> Iterator[torch.nn.Parameter]:
        if not self.empty_sample:
            yield self._sample
        yield from super(Sample, self)._euclidean_parameters(recurse)

    @property
    def sample_transform(self) -> TransformTree:
        r"""
        Default transform used by the sample.
        """
        return self._sample_transform

    @property
    def _sample_transform(self) -> TransformTree:
        def fun():
            return TransformTree(explicit=True,
                                 sample=self.sample[self.idx, :],
                                 default_transform=self._default_sample_transform,
                                 cache_level=self._cache_level)

        return self._get("sample_transform", level_key="sample_transform", fun=fun)

    def transform_input(self, data) -> Union[Tensor, None]:
        r"""
        Apply to value the same transform as on the sample.
        """
        data = utils.castf(data)
        if data is None:
            return None
        return self.sample_transform.apply(x=data)

    def transform_sample_revert(self, data) -> Tensor:
        r"""
        Get back the original value from a projected value, by using the same transform as the sample,
        but in reverse. This is not always feasible, depending on the transform used (normalizations are
        typically not invertible as they are transform which are not bijective).
        """
        return self.sample_transform.revert(data)

    @property
    def hparams_fixed(self) -> dict:
        return {'Input dimension': self.dim_input,
                'Trainable sample': self._sample_trainable,
                'Default sample transforms': self._default_sample_transform,
                **super(Sample, self).hparams_fixed}

    @property
    def hparams_variable(self) -> dict:
        return {**super(Sample, self).hparams_variable}



# coding=utf-8
from ._iterative import (iterative as iterative,
                         iterative_preimage_k as iterative_preimage_k,
                         iterative_preimage_phi as iterative_preimage_phi)
from ._knn import (knn as knn,
                   kernel_knn as kernel_knn)
from ._smoother import (smoother as smoother,
                        kernel_smoother as kernel_smoother)


# coding=utf-8

from __future__ import annotations
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from torch import Tensor as T


def _get_fig_ax(ax: plt.Axes | None = None) -> [plt.Figure, plt.Axes]:
    if ax is None:
        fig = plt.figure()
        ax = fig.gca()
    else:
        assert isinstance(ax, plt.Axes), 'The argument ax is not an instance of matplotlib.Axes.'
        fig = ax.get_figure()
    return fig, ax


def _plot_matrix(m: T,
                 title: str | None = None,
                 x_label: str | None = None,
                 y_label: str | None = None,
                 labels: bool = True,
                 ax: plt.Axes | None = None) -> plt.Figure:
    fig, ax = _get_fig_ax(ax)
    im = ax.imshow(m.data.detach().cpu())
    ax.set_xticks([])
    ax.set_yticks([])

    # title and labels
    if title is not None and labels:
        ax.set_title(title)
    if x_label is None and labels:
        ax.set_xlabel(x_label)
    if y_label is None and labels:
        ax.set_ylabel(y_label)

    # colorbar
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im, cax=cax, orientation='vertical')

    return fig


def plot_K(K: T, labels: bool = False, ax: plt.Axes | None = None) -> plt.Figure:
    r"""
    Plots the kernel matrix.

    :param K: Kernel matrix
    :param labels: If true, the title and axis labels are displayed. Defaults to False.
    :param ax: plt.Axes instance. If None, a new figure/axis pair will be created. Defaults to None.
    :return: The figure

    :type K: torch.Tensor[:,:]
    :type labels: bool, optional
    :type ax: matplotlib.axes.Axes, optional
    :rtype: matplotlib.pyplot.Figure
    """
    return _plot_matrix(K,
                        title='Kernel Matrix',
                        x_label='num_sample',
                        y_label='num_sample',
                        labels=labels,
                        ax=ax)


def plot_K_reconstructed(K_reconstructed: T, labels: bool = False, ax: plt.Axes | None = None) -> plt.Figure:
    r"""
        Plots the reconstructed kernel matrix.

        :param K_reconstructed: The reconstructed kernel matrix.
        :param labels: If true, the title and axis labels are displayed. Defaults to False.
        :param ax: plt.Axes instance. If None, a new figure/axis pair will be created. Defaults to None.
        :return: The figure

        :type K_reconstructed: torch.Tensor[:,:]
        :type labels: bool, optional
        :type ax: matplotlib.axes.Axes, optional
        :rtype: matplotlib.pyplot.Figure
        """
    return _plot_matrix(K_reconstructed,
                        title="Reconstructed Kernel Matrix",
                        x_label='num_sample',
                        y_label='num_sample',
                        labels=labels,
                        ax=ax)


def plot_hidden(hidden: T, labels: bool = False, ax: plt.Axes | None = None) -> plt.Figure:
    r"""
    Plots the hidden variables.

    :param hidden: Hidden variables
    :param labels: If true, the title and axis labels are displayed. Defaults to False.
    :param ax: plt.Axes instance. If None, a new figure/axis pair will be created. Defaults to None.
    :return: The figure

    :type hidden: torch.Tensor[:,:]
    :type labels: bool, optional
    :type ax: matplotlib.axes.Axes, optional
    :rtype: matplotlib.pyplot.Figure
    """
    return _plot_matrix(hidden,
                        title="Hidden Variables",
                        x_label='num_sample',
                        y_label='dim_output',
                        labels=labels,
                        ax=ax)


def plot_hidden_correlation(hidden: T, labels: bool = False, ax: plt.Axes | None = None) -> plt.Figure:
    r"""
    Plots the inner product of the hidden variables.

    :param hidden: Hidden variables
    :param labels: If true, the title and axis labels are displayed. Defaults to False.
    :param ax: plt.Axes instance. If None, a new figure/axis pair will be created. Defaults to None.
    :return: The figure

    :type hidden: torch.Tensor[:,:]
    :type labels: bool, optional
    :type ax: matplotlib.axes.Axes, optional
    :rtype: matplotlib.pyplot.Figure
    """
    return _plot_matrix(hidden.T @ hidden,
                        title="Correlation of the Hidden Variables",
                        x_label='num_sample',
                        y_label='num_sample',
                        labels=labels,
                        ax=ax)


def plot_C(C: T, labels: bool = False, ax: plt.Axes | None = None) -> plt.Figure:
    r"""
    Plots the inner product of the feature vectors.

    :param C: Inner product of the feature vectors.
    :param labels: If true, the title and axis labels are displayed. Defaults to False.
    :param ax: plt.Axes instance. If None, a new figure/axis pair will be created. Defaults to None.
    :return: The figure

    :type C: torch.Tensor[:,:]
    :type labels: bool, optional
    :type ax: matplotlib.axes.Axes, optional
    :rtype: matplotlib.pyplot.Figure
    """
    return _plot_matrix(C,
                        title="Correlation",
                        x_label='dim_feature',
                        y_label='dim_feature',
                        labels=labels,
                        ax=ax)


def plot_Phi(Phi: T, labels: bool = False, ax: plt.Axes | None = None) -> plt.Figure:
    r"""
    Plots a feature vectors matrix.

    :param Phi: Feature vectors
    :param labels: If true, the title and axis labels are displayed. Defaults to False.
    :param ax: plt.Axes instance. If None, a new figure/axis pair will be created. Defaults to None.
    :return: The figure

    :type Phi: torch.Tensor[:,:]
    :type labels: bool, optional
    :type ax: matplotlib.axes.Axes, optional
    :rtype: matplotlib.pyplot.Figure
    """
    return _plot_matrix(Phi,
                        title="Feature Map",
                        x_label='num_sample',
                        y_label='dim_feature',
                        labels=labels,
                        ax=ax)

def plot_W(W: T, labels: bool = False, ax: plt.Axes | None = None) -> plt.Figure:
    r"""
    Plots a weight matrix.

    :param W: Weight matrix
    :param labels: If true, the title and axis labels are displayed. Defaults to False.
    :param ax: plt.Axes instance. If None, a new figure/axis pair will be created. Defaults to None.
    :return: The figure

    :type W: torch.Tensor[:,:]
    :type labels: bool, optional
    :type ax: matplotlib.axes.Axes, optional
    :rtype: matplotlib.pyplot.Figure
    """
    return _plot_matrix(W, title="Weights", x_label='dim_feature', y_label='dim_output', labels=labels, ax=ax)

def plot_sample(sample:T, labels: bool = False, ax: plt.Axes | None = None) -> plt.Figure:
    r"""
    Plots a sample.

    :param sample: Sample
    :param labels: If true, the title and axis labels are displayed. Defaults to False.
    :param ax: plt.Axes instance. If None, a new figure/axis pair will be created. Defaults to None.
    :return: The figure

    :type sample: torch.Tensor[:,:]
    :type labels: bool, optional
    :type ax: matplotlib.axes.Axes, optional
    :rtype: matplotlib.pyplot.Figure
    """
    return _plot_matrix(sample, title="Sample", x_label='num_sample', y_label='dim_intput', labels=labels, ax=ax)

def plot_forward(forward:T, labels: bool = False, ax: plt.Axes | None = None) -> plt.Figure:
    r"""
    Plots the forward pass results.

    :param forward: Matrix containing the forward pass results.
    :param labels: If true, the title and axis labels are displayed. Defaults to False.
    :param ax: plt.Axes instance. If None, a new figure/axis pair will be created. Defaults to None.
    :return: The figure

    :type forward: torch.Tensor[:,:]
    :type labels: bool, optional
    :type ax: matplotlib.axes.Axes, optional
    :rtype: matplotlib.pyplot.Figure
    """
    return _plot_matrix(forward, title="Forward", x_label='num_sample', y_label='dim_output', labels=labels, ax=ax)


def plot_eigenvalues(eigenvalues: T, labels: bool = False, log: bool = True, ax: plt.Axes | None = None) -> plt.Figure:
    r"""
    Bar plot of the eigenvalues.

    :param eigenvalues: Eigenvalues
    :param labels: If true, the title and axis labels are displayed. Defaults to False.
    :param log: If true, the magnitude is in logarithmic scale. En error will be thrown if some values are too close
        to zero or negative. Defaults to True.
    :param ax: plt.Axes instance. If None, a new figure/axis pair will be created. Defaults to None.
    :return: The figure

    :type eigenvalues: torch.Tensor[:]
    :type labels: bool, optional
    :type log: bool, optional
    :type ax: matplotlib.axes.Axes, optional
    :rtype: matplotlib.pyplot.Figure
    """
    fig, ax = _get_fig_ax(ax)
    eigenvalues = eigenvalues.squeeze()
    assert len(eigenvalues.shape) == 1, 'The vals argument must be a vector: each eigenvalue must be a scalar.'

    ax.bar(len(eigenvalues), eigenvalues.detach().cpu())

    if labels:
        ax.set_title('Eigenvalues')
        ax.set_xlabel('dim_output')
        ax.set_ylabel('Magnitude')

    if log:
        ax.set_yscale('log')

    return fig



# coding=utf-8
"""
File containing the feature kernel class.

@author: HENRI DE PLAEN
@copyright: KU LEUVEN
@license: MIT
@date: May 2022
"""

import torch

from .. import utils
from .explicit import Explicit, Kernel
from ._base_kernel import _BaseKernel
from ._factory import factory

@utils.extend_docstring(Kernel)
class Nystrom(Explicit):
    r"""
    Nyström kernel. Constructs an explicit feature map based on the eigendecomposition of any kernel matrix based on
    some sample.

    :param dim: Dimension of the explicit feature map to be constructed. This value cannot exceed the number of sample
        points. During eigendecomposition, very small eigenvalues are also going to be pruned to avoid numerical
        instability. If `None`, the value will be assigned to `num_sample`., defaults to `None`
    :param base_kernel_type: The name of kernel on which the explicit feature map is going to be constructed. Default to
        kerch.DEFAULT_KERNEL_TYPE
    :param base_kernel_transform: Same as kernel_transform but for the base kernel, when using the factory through
        `base_type`. Defaults to [].
    :param \**kwargs: Other arguments for the _Projected kernel (e.g. the bandwidth for an RBF kernel, the degree for a
        polynomial kernel etc.). For the default values, please refer to the requested class in question.
    :param base_kernel: Instead of creating a new kernel on which to use the Nyström method, one can also perform it
        on an existing kernel. In that case, the other _Projected arguments are bypassed., defaults to `None`
    :type dim: int, optional
    :type \**kwargs: dict, optional
    :type base_type: str, optional
    :type base_kernel_transforms: list(str), optional
    :type base_kernel: kerch.kernel.*, optional
    """

    def __init__(self, *args, **kwargs):
        base_kernel_type = kwargs.pop('base_kernel_type', utils.DEFAULT_KERNEL_TYPE)
        assert base_kernel_type.lower() != "nystrom", 'Cannot create a Nyström kernel based on another Nyström ' \
                                                         'kernel.'
        self._base_kernel = None

        k = kwargs.pop('base_kernel', None)
        assert not isinstance(k, str), "base_kernel must be of kernel type (use base_type instead)."
        if k is None:
            # normal case with a kernel created from the factory
            super(Nystrom, self).__init__(*args, **kwargs)

            self._base_kernel = factory(**{**kwargs,
                                               "kernel_type": base_kernel_type,
                                               "kernel_transform": kwargs.pop('base_kernel_transform', [])})
            self._base_kernel.init_sample(sample=self.current_sample_projected, idx_sample=self.idx)
        else:
            # nystromizing some existing kernel
            assert isinstance(k, _BaseKernel), "The provided kernel is not of the kernel class."
            super(Nystrom, self).__init__(**{**kwargs,
                                             "sample": k.sample,
                                             "sample_trainable": k.sample_trainable,
                                             "idx_sample": k.idx})
            self._base_kernel = k
            self._logger.info("Keeping original kernel transform (no overwriting, so base_kernel_transform is "
                           "neglected).")

        self._dim = kwargs.pop('dim', None)
        if self._dim is None:
            self.dim = self._num_total

    @property
    def dim(self):
        """
        Dimension of the explicit feature map.
        """
        return self._dim.cpu().numpy().item(0)

    @dim.setter
    def dim(self, val):
        if self._num_total is not None:
            assert val <= self._num_total, 'Cannot construct an explicit feature map of greater dimension than ' \
                                           'the number of sample points.'
        self._dim = utils.casti(val)
        self._reset_cache(reset_persisting=False)

    @property
    def dim_feature(self) -> int:
        self._compute_decomposition()
        return self.dim

    def __str__(self):
        return "Nystrom kernel"

    @property
    def base_kernel(self):
        r"""
            Kernel on which Nystrom performs the decomposition.
        """
        if self._base_kernel is None:
            raise utils.NotInitializedError(cls=self, message='The base kernel has not been defined yet.')
        return self._base_kernel

    def hparams_fixed(self):
        return {"Kernel": "Nystrom",
                "Base Kernel": self._base_kernel.hparams_fixed['Kernel'],
                **super(Nystrom, self).hparams_fixed}

    def init_sample(self, sample=None, idx_sample=None, prop_sample=None):
        super(Nystrom, self).init_sample(sample=sample, idx_sample=idx_sample, prop_sample=prop_sample)
        if self._base_kernel is not None:
            self._base_kernel.init_sample(sample=self.current_sample_projected, idx_sample=self.idx)

    @torch.no_grad()
    def _compute_decomposition(self):
        if "H" not in self.cache_keys(private=True):
            self._logger.info("Computing the eigendecomposition for the Nystrom kernel.")

            if self._dim is None:
                self.dim = self._num_total

            K = self._base_kernel.K
            lambdas, H = utils.eigs(K, k=self._dim)

            # verify that the decomposed kernel is PSD
            sum_neg = torch.sum(lambdas < 0)
            if sum_neg > 0:
                self._logger.warning(f"The decomposed kernel is not positive semi-definite as it possesses {sum_neg} "
                                  f"negative eigenvalues. These will be discarded, but may prove relevant if their "
                                  f"magnitude is non-negligible.")

            # prune very small eigenvalues if they exist to avoid unstability due to the later inversion
            idx_small = lambdas < utils.EPS
            sum_small = torch.sum(idx_small)
            if sum_small > 0:
                self._logger.warning(
                    f"{sum_small} very small or negative eigenvalues are detected on {self._dim}. "
                    f"To avoid numerical instability, these values are pruned. "
                    f"The new explicit dimension is now {self._dim - sum_small}.")
                keep_idx = torch.logical_not(idx_small)
                lambdas = lambdas[keep_idx]
                H = H[:, keep_idx]
                self._dim -= sum_small

            self._save(key="_nystrom_H", fun=lambda: H, level_key='_nystrom_elements')
            lambdas_sqrt = torch.sqrt(lambdas)
            self._save(key="_nystrom_lambdas_sqrt_inv",
                       fun=lambda: (torch.diag(1 / lambdas_sqrt)).data, level_key='_nystrom_elements')
            self._save(key="_nystrom_sample_phi",
                       fun=lambda: (H @ torch.diag(lambdas_sqrt)).data, level_key='_nystrom_elements')

    def update_sample(self, sample_values, idx_sample=None):
        raise NotImplementedError

    def _explicit_with_none(self, x=None):
        self._compute_decomposition()

        if x is None:
            return self._get(key="_nystrom_sample_phi")

        Kx = self._base_kernel.k(x)
        return Kx @ self._get(key="_nystrom_H") @ self._get(key="_nystrom_lambdas_sqrt_inv")

    def _explicit(self, x):
        # should never happen
        raise Exception("This should never happen, a bug must have occurred.")

    def _explicit_preimage(self, phi) -> torch.Tensor:
        raise utils.ExplicitError(cls=self, message='Explicit pre-image is not possible with the Nystrom kernel.')
